using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Blazedust.CUAM
{

    /// <summary>
    /// This is responsible for keeping track on clothing items per person, what's been loaded and should be kept loaded continously as other clothing presets are loaded/overwriting the items that should be active.
    /// </summary>
    public class ClothingManager
    {

        
        private bool resetAll = false;
        private Dictionary<Atom, PersonClothTracking> trackingItems;
        private int updateIndex = 0;
        private DateTime doubleClickTime;
        private bool supressNextDoubleClickPrompt = false;

        // restoreItems is an "after-fix", hacked in, to get restoring storables to work.
        private List<RestoreClothingStorablesItem> restoreItems = null;

        private const double DOUBLE_CLICK_SECONDS = 1.5D;

        HashSet<string> currentActiveItems;

        CustomUnityAssetManager core;

        public ClothingManager(CustomUnityAssetManager script)
        {
            core = script;
            trackingItems = new Dictionary<Atom, PersonClothTracking>();
            restoreItems = new List<RestoreClothingStorablesItem>();
            doubleClickTime = DateTime.Now;
        }

        /// <summary>
        /// Starts to track a clothing item, making that clothing item continously be loaded if it's deactivated.
        /// </summary>
        public void StartTrackClothingItem(Atom rootAtom, JSONNode clothingNode, DAZClothingItem clothingItem)
        {
            // SuperController.LogMessage("StartTrackClothingItem " + clothingItem.displayName);
            PersonClothTracking clothTracking = null;
            if (!trackingItems.TryGetValue(rootAtom, out clothTracking))
            {
                clothTracking = new PersonClothTracking();
                trackingItems.Add(rootAtom, clothTracking);
            }
            clothTracking.AddTracking(clothingItem, clothingNode);
        }

        /// <summary>
        /// Stop tracking a clothing item.
        /// </summary>
        public void StopTrackClothingItem(Atom rootAtom, string clothingUid)
        {
            // SuperController.LogMessage("StopTrackClothingItem " + clothingUid);
            PersonClothTracking clothTracking = null;
            if (trackingItems.TryGetValue(rootAtom, out clothTracking))
            {
                clothTracking.RemoveTracking(clothingUid);
            }
        }

        /// <summary>
        /// Removes any tracked clothing items currently active.
        /// </summary>
        public void DisableTrackedClothingItems(Atom rootAtom)
        {
            // SuperController.LogMessage("DisableTrackedClothingItems " + rootAtom.uid);
            supressNextDoubleClickPrompt = true;
            PersonClothTracking clothTracking = null;
            if (trackingItems.TryGetValue(rootAtom, out clothTracking))
            {
                DAZCharacterSelector geometry = Helpers.GetDAZCharacterSelector(rootAtom);
                if (geometry != null)
                {
                    foreach (ClothTrackItem item in clothTracking.ClothingItems)
                    {
                        geometry.SetActiveClothingItem(item.ClothingItem, false);
                    }
                }
            }
        }

        /// <summary>
        /// Clears the internal list over tracked clothing items, stop tracking them.
        /// </summary>
        public void ClearTrackClothingItems(Atom rootAtom)
        {
            // SuperController.LogMessage("CLEAR TRACKING CLOTHING ITEMS");
            PersonClothTracking clothTracking = null;
            if (trackingItems.TryGetValue(rootAtom, out clothTracking))
            {
                clothTracking.Dispose();
                trackingItems.Remove(rootAtom);
            }
        }

        /// <summary>
        /// Stop restore clothing storables.
        /// </summary>
        public void StopRestoreClothingStorables(Atom rootAtom, string clothingUid)
        { 
            for (int i = restoreItems.Count - 1; i >= 0; i--)
            {
                RestoreClothingStorablesItem item = restoreItems[i];
                if (item.RootAtom == rootAtom && (item.ClothingItem.uid == clothingUid || clothingUid == ""))
                {
                    restoreItems.RemoveAt(i);
                }
            }
        }

        /// <summary>
        /// Restores clothing storables once the clothing item has loaded its children.
        /// </summary>
        public void StartRestoreClothingStorables(Atom rootAtom, JSONNode clothingNode, DAZClothingItem clothingItem)
        {
            RestoreClothingStorablesItem item = new RestoreClothingStorablesItem(rootAtom, clothingItem, clothingNode);
            restoreItems.Add(item);            
        }

        public void Update(Atom rootAtom)
        {
            if (resetAll)
            {
                // SuperController.LogMessage("CLEAR CLOTHING TRACKING FOR ALL PERSONS");
                foreach(KeyValuePair<Atom, PersonClothTracking> pair in trackingItems)
                {
                    pair.Value.Dispose();
                }
                trackingItems.Clear();
                restoreItems.Clear();
                resetAll = false;
            }

            if (restoreItems.Count > 0)
            {
                for (int k = restoreItems.Count - 1; k >= 0; k --)
                {
                    RestoreClothingStorablesItem item = restoreItems[k];
                    item.Frame++;
                    if (item.Restore())
                    {
                        // SuperController.LogMessage(item.ClothingItem.uid + " restored in " + item.Frame.ToString() + " frames");
                        restoreItems.RemoveAt(k);
                    }
                }
            }

            if (core.options.enableLoadTriggers.val)
            {
                // check more agressively while editing
                int checkFrequency = (SuperController.singleton.activeUI == SuperController.ActiveUI.SelectedOptions && SuperController.singleton.gameMode == SuperController.GameMode.Edit ? 20 : 120);
                if (!SuperController.singleton.isLoading && rootAtom != null && !rootAtom.destroyed && (updateIndex++) > checkFrequency)
                {
                    updateIndex = 1;
                    PersonClothTracking clothTracking = null;
                    if (trackingItems.TryGetValue(rootAtom, out clothTracking))
                    {

                        // Init if not yet done
                        if (!clothTracking.InitDone)
                        {
                            clothTracking.LastActiveItems.Clear();
                            foreach (ClothTrackItem item in clothTracking.ClothingItems)
                            {
                                if (item.ClothingItem.active)
                                {
                                    clothTracking.LastActiveItems.Add(item.ClothingItem.uid);
                                }
                            }
                            clothTracking.InitDone = true;
                            return;
                        }

                        // SuperController.LogMessage("CHECKING " + checkFrequency.ToString());

                        // Check if a clothing preset has been loaded, in that case re-active any clothing items being tracked that might have been disabled by the clothing preset.
                        // We also want the user to be able to manually toggle individual clothing items without this plugin re-activating them (activating or deactivating).

                        // Now, there's no way to see if a clothing preset has been selected explicitly.
                        // But with some clever assumtions, we can figure it out anyway.

                        // If number of clothing items since last time we checked has *changed* with more than one item, we assume a clothing preset has been applied.
                        // *changed* here means the clothing item has gone from active->disabled or disabled->active.
                        // If number of clothing items changes is only one we assume the user has toggled something manually.
                        // If number of clothing items changes is zero nothing has been changed.
                        // The only thing we need to make sure is to keep checking in a frequency that's faster than the user might toggle individual clothing items while in the clothing options menues. 

                        // Problem that comes from the ABOVE approach:
                        // If the user loads the same preset twice, only the tracked clothing items will be disabled on the second load. 
                        // If there are only one tracked clothing item for the person the code will assume it's manually deactivated.
                        // We can "solve" this by having a forced double-click by the user to disable clothing items belonging to a person in the clothing menu. This is OK but not optimal...
                        // Another solution is to always have 2 or more clothing presets bound to the person.
                        // This plugin could for example always bind another invisible clothing item when a clothing preset is saved.
                        // For now - double click!

                        HashSet<string> currentActiveItems = new HashSet<string>(Helpers.GetClothingItems(rootAtom).Select((x) => { return x.uid; }));

                        HashSet<string> disabledItems = new HashSet<string>(clothTracking.LastActiveItems.Except(currentActiveItems));
                        HashSet<string> enabledItems = new HashSet<string>(currentActiveItems.Except(clothTracking.LastActiveItems));
                        bool reloadClothingPresets = ((disabledItems.Count + enabledItems.Count) >= 1);

                        clothTracking.LastActiveItems.Clear();
                        clothTracking.LastActiveItems = currentActiveItems;

                        if (reloadClothingPresets)
                        {
                            DAZCharacterSelector geometry = Helpers.GetDAZCharacterSelector(rootAtom);
                            if (geometry != null)
                            {
                                foreach (ClothTrackItem item in clothTracking.ClothingItems)
                                {
                                    if (!item.ClothingItem.active)
                                    {
                                        // If one - and only one - item was disabled, enforce double-click on that item if it was a tracked clothing item!
                                        if (disabledItems.Count == 1 && enabledItems.Count == 0)
                                        {
                                            if (!disabledItems.Contains(item.ClothingItem.uid))
                                            {
                                                continue;
                                            }
                                            // enforce double-click to disable this item
                                            if (DateTime.Now.Subtract(doubleClickTime).TotalSeconds > DOUBLE_CLICK_SECONDS)
                                            {
                                                if (!supressNextDoubleClickPrompt)
                                                {
                                                    doubleClickTime = DateTime.Now;
                                                    SuperController.LogMessage(string.Format("[CUAManager] Clothing item '{0}' belongs to the person look as a load trigger. Click it again to disable it.", item.ClothingItem.displayName));
                                                }
                                            }
                                            else
                                            {
                                                // user double-clicked now, keep item disabled!
                                                doubleClickTime = DateTime.Now.AddSeconds(-DOUBLE_CLICK_SECONDS);
                                                continue;
                                            }
                                        }

                                        // If one - and only one - item was enabled, don't re-active tracked clothing items. This is OK.
                                        if (disabledItems.Count == 0 && enabledItems.Count == 1)
                                        {
                                            continue;
                                        }
                                    
                                        // Re-load the clothing preset...
                                        core.LoadClothingFromPreset(rootAtom, item.Json, item.ClothingItem, geometry);
                                        clothTracking.LastActiveItems.Add(item.ClothingItem.uid);
                                        // geometry.SetActiveClothingItem(clothingItem, true);
                                    }
                                }
                            }
                        }
                    }

                    supressNextDoubleClickPrompt = false;
                }
            }
        }

        public void Reset()
        {
            resetAll = true;
        }

        public class RestoreClothingStorablesItem
        {
            public Atom RootAtom;
            public DAZClothingItem ClothingItem;
            public JSONNode ClothingNode;
            public int Frame;

            public RestoreClothingStorablesItem(Atom rootAtom, DAZClothingItem clothingItem, JSONNode json)
            {
                RootAtom = rootAtom;
                ClothingItem = clothingItem;
                ClothingNode = json;
                Frame = 0;
            }

            public bool Restore()
            {
                if (RootAtom.destroyed)
                {
                    // root atom destroyed, nothing to restore, done
                    return true;
                }
                // Gather all storables for this clothing item
                Dictionary<string, JSONNode> storeNodes = new Dictionary<string, JSONNode>();
                if (ClothingNode["storables"] != null)
                {
                    foreach (JSONNode storableNode in ClothingNode["storables"].AsArray.Childs)
                    {
                        string storeId = storableNode["id"].Value;
                        storeNodes[storeId] = storableNode;
                    }
                } 
                
                if (storeNodes.Count == 0)
                {
                    // Nothing to restore, done
                    return true;
                }

                bool restored = false;
                // restore all storables for this clothing item
                for (int i = 0; i < ClothingItem.transform.childCount; i++)
                {
                    Transform instance = ClothingItem.transform.GetChild(i);
                    JSONStorable[] componentsInChildren = instance.GetComponentsInChildren<JSONStorable>(includeInactive: true);
                    foreach (JSONStorable js in componentsInChildren)
                    {
                        JSONNode storeNode = null;
                        if (storeNodes.TryGetValue(js.storeId, out storeNode))
                        {
                            js.RestoreFromJSON(storeNode.AsObject);
                            restored = true;
                        }
                    }
                }

                // If Frame count reaches 1000, we can assume it's not going to be loaded... Something wrong with the clothing item or preset.
                if (Frame > 1000 && !restored)
                {
                    SuperController.LogMessage("Clothing item " + ClothingItem.uid + " not restored within 1000 frames - ignoring clothing item.");
                    return true;
                }

                return restored;
            }
        }

        public class ClothTrackItem
        {
            public DAZClothingItem ClothingItem;
            public JSONNode Json;

            public ClothTrackItem(DAZClothingItem clothingItem, JSONNode json)
            {
                ClothingItem = clothingItem;
                Json = json;
            }
        }

        /// <summary>
        /// Helper to track clothing items for a single person
        /// </summary>
        public class PersonClothTracking
        {
            public List<ClothTrackItem> ClothingItems;
            public HashSet<string> LastActiveItems;
            public bool InitDone;

            public PersonClothTracking()
            {
                ClothingItems = new List<ClothTrackItem>();
                LastActiveItems = new HashSet<string>();
                InitDone = false;
            }

            public void AddTracking(DAZClothingItem clothingItem, JSONNode json)
            {
                ClothTrackItem item = new ClothTrackItem(clothingItem, json);
                ClothingItems.Add(item);
                if (clothingItem.active)
                {
                    LastActiveItems.Add(clothingItem.uid);
                }
            }

            public void RemoveTracking(string clothingUid)
            {
                ClothTrackItem item = ClothingItems.Find((x) => { return x.ClothingItem.uid == clothingUid; });
                if (item != null)
                {
                    ClothingItems.Remove(item);
                }
                LastActiveItems.Remove(clothingUid);
            }

            public void Dispose()
            {
                if (ClothingItems != null)
                {
                    ClothingItems.Clear();
                    ClothingItems = null;
                }
                if (LastActiveItems != null)
                {
                    LastActiveItems.Clear();
                    LastActiveItems = null;
                }
            }

        }


    }
}