using System;
using UnityEngine;
using UnityEngine.EventSystems;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Blazedust.CUAM
{
	public class CustomUnityAssetManager : MVRScript 
	{	
		private JSONStorableStringChooser linkedCuaSelector;
		private JSONStorableBool canUpdatePosition;
		private JSONStorableBool canUpdateRotation;
		private JSONStorableBool atomOn;
		private JSONStorableBool atomHidden;
		private JSONStorableBool atomCollision;
		private JSONStorableBool atomPhysicsEnabled;
		private JSONStorableBool atomUseGravity;
		private JSONStorableFloat scaleModifier;
		private JSONStorableString selectedAtomInfo;
		private JSONStorableString savePresetName;
		private JSONStorableStringChooser rootAtomSelector;
		private JSONStorableStringChooser savePresetNameSuggestionSelector;
		private JSONStorableStringChooser loadPresetNameSelector;
		private JSONStorableStringChooser createGenericAssetTypeSelector;
		private JSONStorableStringChooser atomLinkTypeSelector;

		private UIDynamicButton savePresetButton;
		private UIDynamicButton saveClothingPresetButton;
		private UIDynamicButton loadPresetButtonAppend;
		private UIDynamicButton loadPresetButtonReplace;
		private UIDynamicButton clearLinkedCUAButton;
		private UIDynamicButton clearLastLinkedCUAButton;
		private UIDynamicButton refreshPresetsButton;

		private UILoading loading;
		private Atomizer atomizer;
		private SceneAtomTracker atomTracker;
		private UIHelper uiHelper;
		private PosRotSliders posRotSliders;
		private ClothingManager clothingManager;

		private LastAddedAssets lastAddedAssets;
				
		private UIDynamicButton selectedBtn;
		
		private FreeControllerV3 lastSelectedController = null;
		private Atom lastSelectedAtom = null;
		private Atom lastSelectedPersonAtom = null;
		private string lastSelectedAtomUid = "";
		private bool inUpdate = false;
		private bool autoSelectPersonFromSceneQueued = true;
		private SuperController.ActiveUI lastActiveUI = SuperController.ActiveUI.MainMenu;

		private bool pluginActive = false;

		private Queue<UpdateAtomPostCreation> updateAtomPostCreationQueue;
		private Rigidbody preSelectedControllerLinkToRB = null;

		const int FORCE_REMOVE_CLICKS = 4;
		const string TYPE_LINKED_ASSET = "linkedAsset";
		const string TYPE_CLOTHING = "clothing";

		public Options options;
		public Path pluginPath;
		public UIDynamicButton infoTextBtn;
		public Atom rootAtom = null;
		public Atom selectedLinkedAtom = null;

		private const string PLUGIN_NOT_ENABLED = "Plugin not enabled in 'Play Mode' - action ignored.";

		public override void Init() 
		{
			pluginPath = new Path();
			if (!pluginPath.VerifyAndSetup(this)) 
			{
				return;
			}
			pluginActive = true;

			updateAtomPostCreationQueue = new Queue<UpdateAtomPostCreation>();

			options = new Options(this);
			lastAddedAssets = new LastAddedAssets();
			atomTracker = new SceneAtomTracker(this);
			uiHelper = new UIHelper(this);
			loading = new UILoading(this);
			atomizer = new Atomizer(this);
			posRotSliders = new PosRotSliders(this);
			clothingManager = new ClothingManager(this);

			// -- Help button --
			UIDynamicButton btn = uiHelper.CreateButton("Help/Instructions", true);
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
            {
				try {
					Helpers.ShowMainHelpText();
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });
			
			// -- Selector for which atom to work with -- 
			rootAtomSelector = new JSONStorableStringChooser("cuabRootAtom", Helpers.GetRootAtomUids(), "", "Active Person", (string atomUid) => {
				RootAtomSelectedCallback(atomUid);
			});
			UIDynamicPopup popup = uiHelper.CreateScrollablePopup(rootAtomSelector, false);
			rootAtomSelector.popup.onOpenPopupHandlers += () => {
                rootAtomSelector.choices = Helpers.GetRootAtomUids();
            };

			uiHelper.CreateSpacer(false, 30);
			uiHelper.CreateSpacer(true, 80);
			
			// -- Selector for which preset to load -- 
			loadPresetNameSelector = new JSONStorableStringChooser("cuabLoadPresetNameSelector", pluginPath.GetLoadPresetSelectionNames(), "", "Select Preset", (string presetName) => {
				SavePresetSelectedCallback(presetName);
			});
			popup = uiHelper.CreateScrollablePopup(loadPresetNameSelector, false);
			popup.popupPanelHeight = 700f;
			loadPresetNameSelector.popup.onOpenPopupHandlers += () => {
                loadPresetNameSelector.choices = pluginPath.GetLoadPresetSelectionNames();
            };

			// --- Button to append preset (Safe) ---
			loadPresetButtonAppend = uiHelper.CreateButton("Load Preset (Append)", true);
			loadPresetButtonAppend.buttonColor = new Color(0.75f, 0.95f, 0.60f);
			loadPresetButtonAppend.height = 100;
			loadPresetButtonAppend.button.onClick.AddListener(() =>
			{
				try
				{
					// Append is safer to do as it adds atoms (not remove anything). Choke it a bit though.
					if (loading.LoadingInProgress(90))
					{
						return;
					}
					if (!LogicEnabled)
					{
						SuperController.LogMessage(PLUGIN_NOT_ENABLED);
						return;
					}
					LoadSelectedPreset();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			// -- Button to load preset (Replace) - removes existing linked atoms/clothing --
			loadPresetButtonReplace = uiHelper.CreateButton("Load Preset (Replace)\n(make sure no asset is loading!)", true);
			loadPresetButtonReplace.buttonColor = new Color(0.5f, 0.9f, 0.9f);
			loadPresetButtonReplace.height = 100;
			loadPresetButtonReplace.button.onClick.AddListener(() =>
            {
				try {
					if (loading.LoadingInProgress())
					{
						return;
					}
					if (!LogicEnabled)
                    {
						SuperController.LogMessage(PLUGIN_NOT_ENABLED);
						return;
                    }
					LoadSelectedPreset((string type) => {
						try
                        {
							switch (type)
							{
								case TYPE_LINKED_ASSET: 
									RemoveLinkedAssets(rootAtom); 
									break;
								case TYPE_CLOTHING:
									clothingManager.ClearTrackClothingItems(rootAtom);
									Helpers.RemoveAllClothingItems(rootAtom);
									break;
							}
							return true;
						}
						catch (Exception e)
						{
							SuperController.LogError(e.ToString());
							return false;
						}
					});
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });

			// Add a button "Remove Last Added Assets" here... Can be useful when you don't want to remove by selecting assets... :/
			clearLastLinkedCUAButton = uiHelper.CreateButton("Remove Last Added Items (Forced)\n(make sure no asset is loading!)", true);
			clearLastLinkedCUAButton.buttonColor = new Color(1.0f, 0.7f, 1.0f);
			clearLastLinkedCUAButton.height = 80;
			clearLastLinkedCUAButton.button.onClick.AddListener(() =>
			{
				try
				{
					if (loading.LoadingInProgress())
					{
						return;
					}
					if (!LogicEnabled)
					{
						SuperController.LogMessage(PLUGIN_NOT_ENABLED);
						return;
					}
					RemoveLastAddedItems(rootAtom);
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			clearLinkedCUAButton = uiHelper.CreateButton("Remove All Linked Assets", true);
			clearLinkedCUAButton.buttonColor = new Color(1.0f, 0.7f, 0.7f);
			clearLinkedCUAButton.height = 80;
			clearLinkedCUAButton.button.onClick.AddListener(() =>
            {
				try {
					if (loading.LoadingInProgress())
					{
						return;
					}
					if (!LogicEnabled)
					{
						SuperController.LogMessage(PLUGIN_NOT_ENABLED);
						return;
					}
					SetRemoveAllLinkedAssetQuickCounter(removeAllLinkedAssetQuickCounter + 1);
					RemoveLinkedAssets(rootAtom);
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });
			
			refreshPresetsButton = uiHelper.CreateButton("Refresh Preset List", false);
			// refreshPresetsButton.buttonColor = new Color(0.7f, 0.7f, 1.0f);
			refreshPresetsButton.height = 40;
			refreshPresetsButton.button.onClick.AddListener(() =>
            {
				try {
					string preVal = loadPresetNameSelector.val;
					loadPresetNameSelector.choices = pluginPath.GetLoadPresetSelectionNames(true); // force rebuild list
					loadPresetNameSelector.val = (loadPresetNameSelector.choices.Contains(preVal) ? preVal : "");
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });

			uiHelper.CreateSpacer(false, 290);
			uiHelper.CreateSpacer(true, 50);
			
			// -- Button to create preset --
			savePresetButton = uiHelper.CreateButton("Save preset", true);
			savePresetButton.buttonColor = new Color(0.7f, 0.7f, 1.0f);
			savePresetButton.height = 120;
			savePresetButton.button.onClick.AddListener(() =>
            {
				try {
					if (loading.LoadingInProgress())
					{
						return;
					}
					if (!LogicEnabled)
					{
						SuperController.LogMessage(PLUGIN_NOT_ENABLED);
						return;
					}
					SavePreset();
				} catch (Exception e) {
					SuperController.LogError(e.ToString());
				}
            });


			saveClothingPresetButton = uiHelper.CreateButton("Save clothing preset", true);
			saveClothingPresetButton.buttonColor = new Color(0.7f, 0.7f, 1.0f);
			saveClothingPresetButton.height = 120;
			saveClothingPresetButton.button.onClick.AddListener(() =>
			{
				try
				{
					if (loading.LoadingInProgress())
					{
						return;
					}
					if (!LogicEnabled)
					{
						SuperController.LogMessage(PLUGIN_NOT_ENABLED);
						return;
					}
					SaveClothingPreset();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			savePresetName = new JSONStorableString("cuabSavePresetName", "", (string presetName) => {
				savePresetNameSuggestionSelector.valNoCallback = (savePresetNameSuggestionSelector.choices.Contains(presetName) ? presetName : "");
				SavePresetNameCallback(presetName);
			});

			UIDynamicTextField savePresetNameInput = uiHelper.CreateStringTextbox(savePresetName, "Input Save Preset Name...:", false);
			savePresetNameSuggestionSelector = new JSONStorableStringChooser("cuabSavePresetNameSelector", Helpers.GetSavePresetNameSuggestions(rootAtom, CheckIncludeAtomType, ""), "", "...or Select Suggetion", (string presetName) => {
				if (presetName != null && presetName != "")
                {
					savePresetName.val = presetName;
				}
			});
			popup = uiHelper.CreateScrollablePopup(savePresetNameSuggestionSelector, false);
			popup.popupPanelHeight = 700f;
			popup.popup.normalColor = new Color(0.7f, 0.7f, 1.0f);
			savePresetNameSuggestionSelector.popup.onOpenPopupHandlers += () => {
				savePresetNameSuggestionSelector.choices = Helpers.GetSavePresetNameSuggestions(rootAtom, CheckIncludeAtomType, loadPresetNameSelector.val);
			};


			uiHelper.CreateSpacer(false, 50);
			uiHelper.CreateSpacer(true, 22);

			btn = uiHelper.CreateButton("Values for selected linked asset (atom):");
			btn.height = 80;
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					SuperController.singleton.ClearMessages();
					Log("Properties for the current selected and linked asset (atom) to the active person will be available in this section. Here you can align the selected asset relative the linked rigidbody controller and change some basic properties like position, rotation, allow grab, hidden and some more.");
					Log("");
					Log("You can still edit properties by accessing the atom controller's GUI like normally which you must do for certain things like selecting assetbundle file and model. Press 'Select Asset Controller' to quickly select and navigate to the controller's GUI.");
					SuperController.singleton.OpenMessageLogPanel();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			// Selector for current person's linked CustomUnityAsset to control and adjust.
			linkedCuaSelector = new JSONStorableStringChooser("cuabLinkedCuaSelector", Helpers.GetLinkedAtomUids(rootAtom, CheckIncludeAtomType), "", "Selected Asset", (string atomUid) => {
                LinkedCuaSelectedCallback(atomUid);
            });
			popup = uiHelper.CreateScrollablePopup(linkedCuaSelector, false);
			popup.popupPanelHeight = 700f;
			linkedCuaSelector.popup.onOpenPopupHandlers += () => {
                linkedCuaSelector.choices = Helpers.GetLinkedAtomUids(rootAtom, CheckIncludeAtomType);
            };

			infoTextBtn = uiHelper.CreateButton("", true);
			infoTextBtn.height = 80;
			infoTextBtn.buttonColor = new Color(0.5f, 0.6f, 0.8f);
			infoTextBtn.buttonText.text = "";

			btn = uiHelper.CreateButton("Remove Selected Asset (Forced)\n(make sure no asset is loading!)", true);
			btn.height = 100;
			btn.buttonColor = new Color(1.0f, 0.7f, 0.7f);
			btn.button.onClick.AddListener(() => {
				if (loading.LoadingInProgress())
				{
					return;
				}
				RemoveSelectedLinkedCua();
			});

			UIDynamicSlider slider;

			// X,Y,Z sliders for position/rotation
			posRotSliders.Setup(SelectedAssetSliderChanged);

			// Scale modifier
			scaleModifier = new JSONStorableFloat("Scale Modifier", 1.0f, SelectedAssetSliderChanged, 0f, 5f, false);
			slider = uiHelper.CreateSlider(scaleModifier, true);
			slider.valueFormat = "F3";
			slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

			atomLinkTypeSelector = new JSONStorableStringChooser("cuabAtomLinkTypeSelector", options.GetPossibleLinkTypes(), "", "Link Type", SelectedLinkTypeChanged);
			popup = uiHelper.CreateScrollablePopup(atomLinkTypeSelector, true);
			popup.popupPanelHeight = 500f;

			canUpdatePosition = new JSONStorableBool("canUpdatePosition", true, SelectedAssetToggleChanged);
			UIDynamicToggle toggle = uiHelper.CreateToggle(canUpdatePosition);
			toggle.label = "Allow Grab Position";
			
			canUpdateRotation = new JSONStorableBool("canUpdateRotation", true, SelectedAssetToggleChanged);
			toggle = uiHelper.CreateToggle(canUpdateRotation);
			toggle.label = "Allow Grab Rotation";

			atomCollision = new JSONStorableBool("atomCollision", true, SelectedAssetToggleChanged);
			toggle = uiHelper.CreateToggle(atomCollision);
			toggle.label = "Collision";
			toggle.backgroundColor = new Color(0.8f, 0.8f, 1.0f);

			atomUseGravity = new JSONStorableBool("atomUseGravity", true, SelectedAssetToggleChanged);
			toggle = uiHelper.CreateToggle(atomUseGravity);
			toggle.label = "Use Gravity When Position Off";
			toggle.backgroundColor = new Color(0.8f, 0.8f, 1.0f);

			atomPhysicsEnabled = new JSONStorableBool("atomPhysicsEnabled", true, SelectedAssetToggleChanged);
			toggle = uiHelper.CreateToggle(atomPhysicsEnabled);
			toggle.label = "Physics";
			toggle.backgroundColor = new Color(0.8f, 0.8f, 1.0f);

			atomOn = new JSONStorableBool("atomOn", true, SelectedAssetToggleChanged);
			toggle = uiHelper.CreateToggle(atomOn, false);
			toggle.label = "On";
			toggle.backgroundColor = new Color(0.8f, 1.0f, 0.9f);

			atomHidden = new JSONStorableBool("atomHidden", true, SelectedAssetToggleChanged);
			toggle = uiHelper.CreateToggle(atomHidden, false);
			toggle.label = "Hidden";
			toggle.backgroundColor = new Color(0.8f, 1.0f, 0.9f);
			
			btn = uiHelper.CreateButton("Select Asset Controller", false);
			btn.height = 70;
			btn.button.onClick.AddListener(() => {
				if (selectedLinkedAtom != null)
                {
					SuperController.singleton.SelectController(selectedLinkedAtom.uid, "control");
					SuperController.singleton.activeUI = SuperController.ActiveUI.SelectedOptions;
				}
			});

			selectedAtomInfo = new JSONStorableString("cuabSelectedAtomInfoText", "");
			uiHelper.CreateTextField(selectedAtomInfo, true).height = 275;


			uiHelper.CreateSpacer(false, 75);
			uiHelper.CreateSpacer(true, 255);

			btn = uiHelper.CreateButton("Create linked atom (generic): ", false);
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					SuperController.singleton.ClearMessages();
					Log("Creates a linked atom to the predefined rigidbody link for the selected person with some preset values for position and rotation offsets.");
					SuperController.singleton.OpenMessageLogPanel();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			// Selector for current person's linked CustomUnityAsset to control and adjust.
			createGenericAssetTypeSelector = new JSONStorableStringChooser("cuabCreateGenericAssetTypeSelector", options.GetIncludedAtomTypes(), "CustomUnityAsset", "Atom Type");
			popup = uiHelper.CreateScrollablePopup(createGenericAssetTypeSelector, false);
			popup.popupPanelHeight = 700f;
			createGenericAssetTypeSelector.popup.onOpenPopupHandlers += () => {
				createGenericAssetTypeSelector.choices = options.GetIncludedAtomTypes();
			};

			CreateLinkableAssets(
				new LinkableAsset("head (top)", "head", "CUA_headTop", new Vector3(0.0f, 0.093f, 0.03f), new Vector3(0f, 0f, 0f)),
				new LinkableAsset("head (back)", "head", "CUA_headBack", new Vector3(0.0f, 0.042f, -0.022f), new Vector3(340f, 0f, 0f)),
				new LinkableAsset("head (mouth)", "head", "CUA_headMouth", new Vector3(0.0f, 0.0f, 0.1f), new Vector3(0f, 0f, 0f), guiRight:true),
				new LinkableAsset("head (lEar)", "head", "CUA_headLEar", new Vector3(-0.075f, 0.017f, 0.004f), new Vector3(0f, 0f, 0f)),
				new LinkableAsset("head (rEar)", "head", "CUA_headREar", new Vector3(0.075f, 0.017f, 0.004f), new Vector3(0f, 0f, 0f), guiRight:true),
				new LinkableAsset("neck", "neck", "CUA_neck", new Vector3(0.0f, 0.014f, 0.04f), new Vector3(0f, 0f, 0f)),
				new LinkableAsset("lShldr (lShoulder)", "lShldr", "CUA_lShldr", new Vector3(0.0f, 0.03f, 0.0f), new Vector3(0f, 0f, 0f)),
				new LinkableAsset("rShldr (rShoulder)", "rShldr", "CUA_rShldr", new Vector3(0.0f, 0.03f, 0.0f), new Vector3(0f, 0f, 0f), guiRight:true),
				new LinkableAsset("lForeArm", "lForeArm", "CUA_lForeArm", new Vector3(0.03f, 0.0f, -0.1f), new Vector3(30f, 0f, 0f)),
				new LinkableAsset("rForeArm", "rForeArm", "CUA_rForeArm", new Vector3(-0.03f, 0.0f, -0.1f), new Vector3(30f, 0f, 0f), guiRight:true),
				new LinkableAsset("lHand", "lHand", "CUA_lHand", new Vector3(-0.073f, -0.03f, 0.0f), new Vector3(90f, 332f, 280f)),
				new LinkableAsset("rHand", "rHand", "CUA_rHand", new Vector3(0.073f, -0.03f, 0.0f), new Vector3(90f, 16f, 70f), guiRight:true),
				new LinkableAsset("chest (chest)", "chest", "CUA_chest", new Vector3(0.00f, 0.1f, 0.08f), new Vector3(0f, 0f, 0f)),
				new LinkableAsset("chest (back)", "chest", "CUA_chestBack", new Vector3(0.0f, 0.050f, -0.033f), new Vector3(0f, 0f, 0f), guiRight: true),
				new LinkableAsset("abdomen (back)", "abdomen", "CUA_abdomenBack", new Vector3(0.0f, 0.01f, -0.055f), new Vector3(0f, 180f, 0f)),
				new LinkableAsset("abdomen2 (back)", "abdomen2", "CUA_abdomen2Back", new Vector3(0.0f, 0.016f, -0.060f), new Vector3(0f, 180f, 0f), guiRight: true),
				new LinkableAsset("pelvis", "pelvis", "CUA_pelvis", new Vector3(0.0f, -0.050f, 0.100f), new Vector3(0f, 0f, 0f)),
				new LinkableAsset("hip", "hip", "CUA_hip", new Vector3(0.0f, -0.07f, 0.03f), new Vector3(0f, 0f, 0f), guiRight: true),
				new LinkableAsset("lShin (lKnee)", "lShin", "CUA_lShin", new Vector3(0.0f, 0.03f, 0.0f), new Vector3(0f, 0f, 0f)),
				new LinkableAsset("rShin (rKnee)", "rShin", "CUA_rShin", new Vector3(0.0f, 0.03f, 0.0f), new Vector3(0f, 0f, 0f), guiRight:true),
				new LinkableAsset("lFoot (lFoot)", "lFoot", "CUA_lFoot", new Vector3(0.0f, 0.03f, 0.0f), new Vector3(0f, 0f, 0f)),
				new LinkableAsset("rFoot (rFoot)", "rFoot", "CUA_rFoot", new Vector3(0.0f, 0.03f, 0.0f), new Vector3(0f, 0f, 0f), guiRight:true)
				);


			// Spaces
			uiHelper.CreateSpacer(false, 50);
			uiHelper.CreateSpacer(true, 50);
			
			options.InitGui();
			options.LoadOptions();
			
			// post init stuff...
			SavePresetNameCallback("");
			SetRemoveAllLinkedAssetQuickCounter(0);
		}

		public bool CheckIncludeAtomType(string atomType)
        {
			return options.CheckIncludeAtomType(atomType);
		}

		private int removeAllLinkedAssetQuickCounter = 0;
		private float removeAllLinkedAssetLastClickRealtimeSinceStartup = 0f;
		
		private void SetRemoveAllLinkedAssetQuickCounter(int counter)
        {
			if (counter > 0)
            {
				removeAllLinkedAssetQuickCounter = counter;
				removeAllLinkedAssetLastClickRealtimeSinceStartup = Time.realtimeSinceStartup;
				if (removeAllLinkedAssetQuickCounter > 1)
				{
					clearLinkedCUAButton.label = string.Format("Remove All Linked Assets (Forced)\n(force remove everything in {0} clicks)", (FORCE_REMOVE_CLICKS - removeAllLinkedAssetQuickCounter).ToString());
				}
			} 
			else
            {
				removeAllLinkedAssetQuickCounter = 0;
				clearLinkedCUAButton.label = "Remove All Linked Assets (Filtered)\n(make sure no asset is loading!)";
			}
        }
		
		public void AutoLoaderTriggered(Atom parentAtom, int loadMode, List<string> presetNamesToLoadOnChange, List<string> presetNamesToUnloadOnChange)
        {
			// Remove linked assets created this "scene edit session" if smartRemovalOfPresets is NOT activated. 
			// This is legacy how it worked (just replace everything, don't be smart!)
			if (!options.smartRemovalOfPresets.val && loadMode == 1)
			{
				RemoveLinkedAssets(parentAtom);
			}

			// Disable current tracked clothing items when auto load triggers are changed.
			// Need to do this as there's some delay between selecting a new look, starting to load it and triggers are eventually updated.
			// If the clothingManager gets an update between, it's gonna enable clothing items not for this look.
			// An easy way to get this to work together with current plugin framework is just to disable all tracked clothing items and let them be re-enabled. Only occurs when auto-load-triggers are updated!
			clothingManager.DisableTrackedClothingItems(parentAtom);
			
			// Remove linked scene assets if configured to do so, 1.2 feature in this plugin.
			if (presetNamesToUnloadOnChange != null)
            {
				// SuperController.LogMessage("presetNamesToUnloadOnChange: " + string.Join(", ", presetNamesToUnloadOnChange.ToArray()));
				// Load presets in memory that was removed (but don't create atoms) and figure out if any scene atoms are the same as these presets.
				foreach(string presetName in presetNamesToUnloadOnChange)
                {
					string file = pluginPath.GetFileFromPresetSelection(presetName);
					if (file != "") 
					{
						List<PresetItemMetaInfo> presetInfos = Helpers.GetPresetMetaInfo(file, CheckIncludeAtomType);
						if (presetInfos != null)
                        {
							Helpers.RemoveScenePreset(parentAtom, presetInfos, CheckIncludeAtomType);

							// Stop tracking clothing items no longer needed to be tracked
							IEnumerable<PresetItemMetaInfo> clothingPresetInfos = presetInfos.Where((x) => { return x.isClothingPreset; });
							if (clothingPresetInfos.Any())
                            {
								foreach (PresetItemMetaInfo presetInfo in clothingPresetInfos)
								{
									clothingManager.StopTrackClothingItem(parentAtom, presetInfo.clothingUid);
									clothingManager.StopRestoreClothingStorables(parentAtom, presetInfo.clothingUid);
								}
							}

						}
					}
				}
            }
			
			// Prepare and start loading new presets
			if (presetNamesToLoadOnChange != null)
			{
				foreach (string presetName in presetNamesToLoadOnChange)
				{
					string file = pluginPath.GetFileFromPresetSelection(presetName);
					if (file != "")
					{
						LoadPresetFromFile(file, parentAtom, true, (string type) => {
							
							// loadMode 2 : CLOTHING ONLY, Only allow clothing to load.
							if (loadMode == 2 && type != TYPE_CLOTHING)
								return false;

							return true;
						});
					}
				}
			}

			// update save preset button
			SavePresetNameCallback(savePresetName.val);
		}

		/// <summary>
		/// Gets if the plugin logic is enabled.
		/// GameMode is Edit or user has opted in to run this plugin while GameMode is Play.
		/// </summary>
		public bool LogicEnabled { 
			get {
				return SuperController.singleton.gameMode == SuperController.GameMode.Edit || options.enablePlayModeUpdates.val;
			} 
		}

		void Update()
        {
			if (!pluginActive)
				return;

			inUpdate = true;
			int atPart = 0;
			try
			{
				if (removeAllLinkedAssetQuickCounter > 0 && removeAllLinkedAssetLastClickRealtimeSinceStartup + 0.75f < Time.realtimeSinceStartup)
                {
					SetRemoveAllLinkedAssetQuickCounter(0);
				}

				if (SuperController.singleton.isLoading) 
				{
					atomTracker.Reset();
					options.Reset();
					clothingManager.Reset();
					autoSelectPersonFromSceneQueued = true;
					return;
				}

				if (!LogicEnabled)
				{
					return;
				}

				if (atomizer.Update())
				{
					return;
				}

				if (cuaPackages.Count > 0)
                {
					for (int i = 0; i < cuaPackages.Count; i ++)
                    {
						if (cuaPackages[i].Update())
						{
							if (cuaPackages[i].IsCompleted())
							{
								cuaPackages.RemoveAt(i);
								i--;
							}
							break;
						}
					}
				}

				atomTracker.Update();

				atPart = 1;
				if (rootAtom != null && rootAtom.destroyed) 
				{
					rootAtom = null;
					selectedLinkedAtom = null;
				}
				if (rootAtom == null && rootAtomSelector.val != "")
				{
					rootAtomSelector.valNoCallback = "";
					rootAtomSelector.choices = new List<string>();
					savePresetName.valNoCallback = "";
					savePresetNameSuggestionSelector.valNoCallback = "";
				}
				if (selectedLinkedAtom == null && linkedCuaSelector.val != "")
                {
					RemoveSelectedLinkedCua();
				}

				atPart = 2;
				Atom selectedAtom = SuperController.singleton.GetSelectedAtom();
				if (selectedAtom != null && lastSelectedPersonAtom != selectedAtom && selectedAtom.type == "Person")
				{
					options.PersonDeselected(lastSelectedPersonAtom);
					lastSelectedPersonAtom = selectedAtom;
					options.PersonSelected(lastSelectedPersonAtom);
				}

				if (loading.PostDelay())
				{
					// don't process anything else this frame...
					return;
                }

				if (updateAtomPostCreationQueue.Count > 0)
                {
					UpdateAtomPostCreation data = updateAtomPostCreationQueue.Peek();
					data.delayFrames--;
					if (data.delayFrames <= 0)
                    {
						updateAtomPostCreationQueue.Dequeue();
						if (!data.apAtom.destroyed)
						{
							data.apAtom.mainController.currentPositionState = data.posState;
							data.apAtom.mainController.currentRotationState = data.rotState;
							data.apAtom.collisionEnabledJSON.val = data.collisionEnabled;
							data.apAtom.GetStorableByID("control").SetBoolParamValue("physicsEnabled", data.physicsEnabled);
							updateOffsetsFromSelectedAtomQueued = true;
						}
					}
				}
				
				atPart = 3;
				FreeControllerV3 selectedController = SuperController.singleton.GetSelectedController();
				if (lastSelectedAtom != selectedAtom)
				{
					lastSelectedAtom = selectedAtom;
					lastSelectedController = selectedController;
					if (selectedAtom != null) 
					{
						lastSelectedAtomUid = selectedAtom.uid;
						// selectedAtom has changed
						if (selectedAtom.type == "Person" && options.autoSelectPersonFromScene.val) 
						{
							SelectRootAtom(selectedAtom.uid);
						} 
						else if (CheckIncludeAtomType(selectedAtom.type) && options.autoSelectAssetController.val)
                        {
							if (options.autoSelectPersonFromScene.val)
                            {
								// make sure rootAtom is correctly selected, if this atom is linked!
								LinkToInfo linkToInfo = Helpers.GetLinkToInfo(selectedAtom);
								if (linkToInfo != null && linkToInfo.linkToAtomUid != "" && (rootAtom == null || rootAtom.uid != linkToInfo.linkToAtomUid))
                                {
									SelectRootAtom(linkToInfo.linkToAtomUid);
                                }
							}
							// Check if it's within current rootAtom's linked atoms.
						 	List<string> cuaOptions = Helpers.GetLinkedAtomUids(rootAtom, CheckIncludeAtomType);
							if (cuaOptions.Contains(selectedAtom.uid))
                            {
								linkedCuaSelector.choices = cuaOptions;
								linkedCuaSelector.val = selectedAtom.uid;
								updateOffsetsFromSelectedAtomQueued = true;
							}
						}

						// Log("Selected atom updated to " + selectedAtom.uid + " of type " + selectedAtom.type);
						// Log("Selected controller updated to " + selectedController.transform.name);
					} else {
						lastSelectedAtomUid = "";
					}
				} 
				else if (lastSelectedController != selectedController)
				{
					lastSelectedController = selectedController;
					if (selectedController != null) 
					{
						// Log("Selected controller updated to " + selectedController.transform.name);
					}
				}
			
				atPart = 4;
			
				if (selectedAtom != null && lastSelectedAtomUid != selectedAtom.uid) 
				{
					// Log("Selected atom name changed from '" + lastSelectedAtomUid + "' to '" + selectedAtom.uid + "'");
					// rename happened. Reselect atom if it was selected as the name has been updated.
					if (selectedAtom.type == "Person" && rootAtomSelector.val == lastSelectedAtomUid)
					{
						SelectRootAtom(selectedAtom.uid);
					}
					if (selectedAtom.type != "Person" && linkedCuaSelector.val == lastSelectedAtomUid)
					{
						linkedCuaSelector.valNoCallback = selectedAtom.uid;
					}

					RefreshSavePresetNames();
					lastSelectedAtomUid = selectedAtom.uid;
				}
			
				atPart = 5;

				// select first best Person in the scene automatically when plugin boots up.
				if (autoSelectPersonFromSceneQueued) 
				{
					autoSelectPersonFromSceneQueued = false;
					if (options.autoSelectPersonFromScene.val && rootAtomSelector.val == "")
					{
						List<string> rootAtomUids = Helpers.GetRootAtomUids();
						if (rootAtomUids.Count > 0)
						{
							SelectRootAtom(rootAtomUids[0]);
						}
					}
				}

				atPart = 6;

				clothingManager.Update(lastSelectedPersonAtom);

				options.Update(lastSelectedPersonAtom);
			
				if (updateAtomToLinkQueued)
				{
					atPart = 7;
					updateAtomToLinkQueued = false;
					try {
						if (selectedLinkedAtom != null)
						{
							atPart = 8;
							UpdateAtomToLink(selectedLinkedAtom, rootAtom, selectedLinkedAtom.mainController.linkToRB, Vector3.zero, Vector3.zero, applyPluginValues: true);
						}
					} catch (Exception e) {
						SuperController.LogError(e.ToString());
					}
				} else {
					atPart = 9;
					if (selectedLinkedAtom != null && selectedLinkedAtom.mainController.linkToRB != null)
					{
						atPart = 10;
						if (selectedLinkedAtom.mainController.linkToRB.name == "SelectionHandles" || selectedLinkedAtom.mainController.linkToRB.name == "MouseGrab" || SuperController.singleton.activeUI != SuperController.ActiveUI.MainMenu)
						{
							updateOffsetsFromSelectedAtomQueued = true;
							if (SuperController.singleton.activeUI == SuperController.ActiveUI.MainMenu && selectedController == selectedLinkedAtom.mainController)
							{
								posRotSliders.UpdateOffsetsPosRotFromAtom(selectedLinkedAtom, rootAtom, preSelectedControllerLinkToRB);
							}
						}
						else if (updateOffsetsFromSelectedAtomQueued)
						{
							preSelectedControllerLinkToRB = selectedLinkedAtom.mainController.linkToRB;
							UpdateOffsetsFromAtom(selectedLinkedAtom, rootAtom, selectedLinkedAtom.mainController.linkToRB);
							updateOffsetsFromSelectedAtomQueued = false;
							// updates save preset button in case this is a new linked atom (from empty to non-empty linked atoms).
							SavePresetNameCallback(savePresetName.val);
						}
						//else if (SuperController.singleton.activeUI == SuperController.ActiveUI.MainMenu)
						//{
						//	posRotSliders.UpdateOffsetsPosRotFromAtom(selectedLinkedAtom, rootAtom, selectedLinkedAtom.mainController.linkToRB, options.LiveUpdateLink(selectedLinkedAtom.mainController.currentPositionState), options.LiveUpdateLink(selectedLinkedAtom.mainController.currentRotationState));
						//}
					}
					else if (selectedLinkedAtom != null)
                    {
						// Log("selectedLinkedAtom No Longer Linked!");
						linkedCuaSelector.val = "";
						// updates save preset button
						SavePresetNameCallback(savePresetName.val);
						updateOffsetsFromSelectedAtomQueued = false;
					}
					else 
					{
						updateOffsetsFromSelectedAtomQueued = false;
						if (lastActiveUI != SuperController.singleton.activeUI)
						{
							lastActiveUI = SuperController.singleton.activeUI;
							if (lastActiveUI == SuperController.ActiveUI.MainMenu)
                            {
								// main menu focused, perform updates in plugin GUI when it's focused here.

								// updates save preset button
								SavePresetNameCallback(savePresetName.val);
							}
						}
					}

					if (selectedLinkedAtomUpdatedDelay > 0)
                    {
						selectedLinkedAtomUpdatedDelay--;
						if (selectedLinkedAtom != null && selectedLinkedAtomUpdatedDelay <= 0 && options.autoSelectAssetController.val)
						{
							// Only auto-select it if we're in the MainMenu (plugin-session-tab, or we can't really select it without it sometimes taking focus from what the user is doing directly in the scene)
							if (SuperController.singleton.activeUI == SuperController.ActiveUI.MainMenu)
							{
								SuperController.singleton.SelectController(selectedLinkedAtom.uid, "control");
								SuperController.singleton.activeUI = SuperController.ActiveUI.MainMenu;
							}
						}
					} 
					else if (selectedLinkedAtom == null && selectedAtomInfo.val != "")
                    {
						// Clear info-box over selected linked atom if none exist. Could reset other gui-components too but that's not required.
						selectedAtomInfo.val = "";
					}
				}

			} catch (Exception e) {
				SuperController.LogError("Error at update " + atPart.ToString());
				SuperController.LogError(e.ToString());
			} finally {
				inUpdate = false;
			}
		}

		void OnDestroy()
		{
			posRotSliders.Destroy();
			posRotSliders = null;
			options.Destroy();
			options = null;
			uiHelper.Destroy();
			uiHelper = null;
			loading.Destroy();
			loading = null;
			atomizer.Destroy();
			atomizer = null;
			atomTracker.Destroy();
			atomTracker = null;
			pluginPath = null;
			rootAtom = null;
			selectedLinkedAtom = null;
			lastLoadedjn = null;
		}

		int selectedLinkedAtomUpdatedDelay = 0;
		bool updateOffsetsFromSelectedAtomQueued = false;
		bool updateAtomToLinkQueued = false;
		protected void SelectedAssetSliderChanged(JSONStorableFloat jf)
		{
			if (!updateAtomToLinkQueued && !updateOffsetsFromSelectedAtomQueued) 
			{
				updateAtomToLinkQueued = true;
			}
		}
		protected void SelectedAssetToggleChanged(bool val)
		{
			if (!updateAtomToLinkQueued && !updateOffsetsFromSelectedAtomQueued) 
			{
				updateAtomToLinkQueued = true;
			}
		}
		protected void SelectedLinkTypeChanged(string linkType)
        {
			if (!updateAtomToLinkQueued && !updateOffsetsFromSelectedAtomQueued)
			{
				updateAtomToLinkQueued = true;
			}
		}

		protected void RootAtomSelectedCallback(string uid)
		{
			SelectRootAtom(uid);
			RefreshSavePresetNames();
		}
		protected void SavePresetSelectedCallback(string presetName)
		{
			RefreshSavePresetNames();
		}
		protected void SavePresetNameCallback(string presetName)
		{
			if (presetName == "")
			{
				savePresetButton.buttonColor = new Color(1.0f, 1.0f, 0.7f);
				savePresetButton.buttonText.text = "Save Linked Atoms Preset\nSelect name before saving preset for all current linked atoms (assets)";

				saveClothingPresetButton.buttonColor = new Color(1.0f, 1.0f, 0.7f);
				saveClothingPresetButton.buttonText.text = "Save Clothing Preset\nSelect name before saving clothing preset";
				
				return;
			}

			int linkedAssetCount = Helpers.GetChildAtoms(rootAtom, CheckIncludeAtomType).Count();
			int clothingCount = Helpers.GetClothingItems(rootAtom).Count();
			PresetFileInfo presetInfo = (linkedAssetCount != 0 || clothingCount != 0 ? pluginPath.GetSavePresetFileInfo(presetName) : null);

			if (linkedAssetCount == 0)
            {
				savePresetButton.buttonColor = new Color(1.0f, 1.0f, 0.7f);
				savePresetButton.buttonText.text = "Save Linked Atoms Preset\nNo linked atoms (assets) for selected person. Nothing to save";
			}
			else
			{
				if (presetInfo.isValid && presetInfo.fileExist)
				{
					savePresetButton.buttonColor = new Color(1.0f, 0.7f, 0.7f);
					savePresetButton.buttonText.text = string.Format("Save Linked Atoms Preset\nOverwrite existing preset for all ({0}) linked atoms (assets)", linkedAssetCount);
				}
				else
				{
					savePresetButton.buttonColor = new Color(0.7f, 0.7f, 1.0f);
					savePresetButton.buttonText.text = string.Format("Save Linked Atoms Preset\nSave preset for all ({0}) linked atoms (assets)", linkedAssetCount);
				}
			}

			if (clothingCount == 0)
            {
				saveClothingPresetButton.buttonColor = new Color(1.0f, 1.0f, 0.7f);
				saveClothingPresetButton.buttonText.text = "Save Clothing Preset\nNo clothing active for selected person. Nothing to save";
			} 
			else
            {
				if (presetInfo.isValid && presetInfo.fileExist)
				{
					saveClothingPresetButton.buttonColor = new Color(1.0f, 0.7f, 0.7f);
					saveClothingPresetButton.buttonText.text = string.Format("Save Clothing Preset\nOverwrite existing preset for all ({0}) clothing items", clothingCount);
				}
				else
				{
					saveClothingPresetButton.buttonColor = new Color(0.7f, 0.7f, 1.0f);
					saveClothingPresetButton.buttonText.text = string.Format("Save Clothing Preset\nSave preset for all ({0}) clothing items", clothingCount);
				}
			}

		}
		protected void LinkedCuaSelectedCallback(string linkedUid)
		{
			if (linkedUid == null || linkedUid == "")
			{
				selectedLinkedAtom = null;
				return;
            }
			if (rootAtom == null || rootAtom.destroyed)
			{
				Log("Select a person first.");
				return;
			}

			selectedLinkedAtom = SuperController.singleton.GetAtomByUid(linkedUid);
			selectedLinkedAtomUpdatedDelay = 25;
			updateOffsetsFromSelectedAtomQueued = true;
		}
		
		private void RemoveSelectedLinkedCua()
		{
            if (selectedLinkedAtom != null && !selectedLinkedAtom.destroyed)
            {
                selectedLinkedAtom.Remove();
                selectedLinkedAtom = null;
			}
			linkedCuaSelector.val = "";
			linkedCuaSelector.choices = Helpers.GetLinkedAtomUids(rootAtom, CheckIncludeAtomType);
			RefreshSavePresetNames();
		}

		// Refresh save preset names so you don't by mistake select the wrong one no longer available
		private void RefreshSavePresetNames()
		{
			string currentSavePresetName = savePresetName.val;
			List<string> newChoices = Helpers.GetSavePresetNameSuggestions(rootAtom, CheckIncludeAtomType, loadPresetNameSelector.val);
			if (savePresetNameSuggestionSelector.val != null && savePresetNameSuggestionSelector.val != null && 
				!newChoices.Contains(savePresetNameSuggestionSelector.val) && savePresetName.val == savePresetNameSuggestionSelector.val)
			{
				currentSavePresetName = "";
			}

			savePresetNameSuggestionSelector.choices = newChoices;
			if (savePresetNameSuggestionSelector.val != null && savePresetNameSuggestionSelector.val != "" && !newChoices.Contains(savePresetNameSuggestionSelector.val)) 
			{	
				savePresetNameSuggestionSelector.valNoCallback = "";
			}

			savePresetName.val = currentSavePresetName;
			SavePresetNameCallback(savePresetName.val); // force callback in case savePresetName.val == currentSavePresetName...
		}
		
		private JSONNode lastLoadedjn = null;

		/// <summary>
		/// Loads the current selected preset for the current selected person.
		/// </summary>
		/// <param name="preLoadCallback">Option callback to run before preset is loaded, with the type of preset as param1</param>
		private void LoadSelectedPreset(Func<string, bool> preLoadCallback = null)
		{
			if (rootAtom == null || rootAtom.destroyed) {
				Log("Select a person first.");
				return;
			}
			
			if (loadPresetNameSelector.val == "") {
				Log("Select a preset first.");
				return;
			}
			
			string file = pluginPath.GetFileFromPresetSelection(loadPresetNameSelector.val);
			if (file == "") {
				Log("Preset file mapping not found, try to refresh preset list.");
				return;
			}

			LoadPresetFromFile(file, rootAtom, false, preLoadCallback);
			options.SuggestLoadPreset(loadPresetNameSelector.val);
		}

		private List<PresetCuaPackage> cuaPackages = new List<PresetCuaPackage>();

		/// <summary>
		/// Loads a preset from file to selected atom
		/// </summary>
		/// <param name="file"></param>
        /// <param name="atom"></param>
        /// <param name="addAsClothTrackingItem">If a clothing preset should be kept tracked and loaded on the person as clothing items are changed.</param>
        /// <param name="preLoadCallback">Option callback to run before preset is loaded, with the type of preset as param1</param>
		private void LoadPresetFromFile(string file, Atom atom, bool addAsClothTrackingItem, Func<string, bool> preLoadCallback = null)
        {
			try
			{
				JSONNode jn = SuperController.singleton.LoadJSON(file);
				JSONClass jc = jn.AsObject;

				lastLoadedjn = jn; // keep reference to avoid GC collecting this (even if unlikely)

				// read what type of preset is available
				if (jn["items"] != null)
				{
					if (preLoadCallback != null && !preLoadCallback.Invoke(TYPE_LINKED_ASSET))
                    {
						return;
					}

					int expectedCount = jn["items"].AsArray.Count;
					if (expectedCount <= 0)
						return;

					lastAddedAssets.Clear(atom);

					PresetCuaPackage cuaPackage = new PresetCuaPackage(expectedCount, options.updateScriptAtomUids.val, atom.uid);
					cuaPackages.Add(cuaPackage);

					foreach (JSONNode itemNode in jn["items"].AsArray.Childs)
					{
						CreateCUAFromPreset(atom, itemNode, cuaPackage);
					}

					// updates save preset
					if (rootAtom == atom)
					{
						RefreshSavePresetNames();
						savePresetName.val = loadPresetNameSelector.val;
					}
				}

				if (jn["clothing"] != null)
                {
					if (preLoadCallback != null && !preLoadCallback.Invoke(TYPE_CLOTHING))
					{
						return;
					}

					int expectedCount = jn["clothing"].AsArray.Count;
					if (expectedCount <= 0)
						return;

					lastAddedAssets.Clear(atom);

					LoadAllClothingFromPreset(atom, jn, addAsClothTrackingItem);

					// updates save preset
					if (rootAtom == atom)
					{
						RefreshSavePresetNames();
						savePresetName.val = loadPresetNameSelector.val;
					}
				}
			}
			catch (Exception e)
			{
				SuperController.LogError("Error loading file: " + file);
				SuperController.LogError(e.ToString());
			}
		}

		/// <param name="addAsClothTrackingItem">If a clothing preset should be kept tracked and loaded on the person as clothing items are changed.</param>
		private void LoadAllClothingFromPreset(Atom rootAtom, JSONNode jn, bool addAsClothTrackingItem)
        {
			// helper to load all clothing items in the preset file
			Dictionary<string, DAZClothingItem> clothingItems = Helpers.GetAllClothingItems(rootAtom);
			DAZCharacterSelector geometry = Helpers.GetDAZCharacterSelector(rootAtom);
			foreach (JSONNode clothingNode in jn["clothing"].AsArray.Childs)
			{
				string clothingUid = clothingNode["clothingUid"].Value;
				DAZClothingItem clothingItem = null;
				if (clothingItems.TryGetValue(clothingUid, out clothingItem))
				{
					LoadClothingFromPreset(rootAtom, clothingNode, clothingItem, geometry);
					if (addAsClothTrackingItem)
					{
						clothingManager.StartTrackClothingItem(rootAtom, clothingNode, clothingItem);
					} 
					else
                    {
						lastAddedAssets.AddClothing(rootAtom, clothingItem);
					}
				}
			}
		}

		public void LoadClothingFromPreset(Atom rootAtom, JSONNode clothingNode, DAZClothingItem clothingItem, DAZCharacterSelector geometry)
		{
			// string clothingUid = clothingNode["clothingUid"].Value;

			// Activate the clothing item this frame.
			geometry.SetActiveClothingItem(clothingItem, true);

			// Restore storables. Can't be done until clothing item is fully loaded and child components exist.
			// This will happen sometime in future frames.
			// The first frame the clothing item gets activated it doesn't contain all the child components yet.
			// I figured this out late - so this code is a late addition and a bit messy but works.
			clothingManager.StartRestoreClothingStorables(rootAtom, clothingNode, clothingItem);
		}

		private void CreateCUAFromPreset(Atom rootAtom, JSONNode cuaNode, PresetCuaPackage presetPackage)
		{
			// Log("Creating " + cuaNode["cua"].Value + " from preset");
			string atomId = cuaNode["cua"].Value;
			JSONNode jnAtom = cuaNode["atoms"].AsArray.Childs.FirstOrDefault((x) => { 
				if (!CheckIncludeAtomType(x["type"].Value))
                {
					SuperController.LogMessage(string.Format("Atom type '{0}' in preset file not configured to be handled by the CUAManager plugin. Ignoring atom type.", x["type"].Value));
					return false;
				}
				return true;
			});
			if (jnAtom == null)
			{
				if (presetPackage != null)
				{
					presetPackage.OneNotFound();
				}
				return;
			}

			JSONNode jnCuabLinkTo = jnAtom["storables"].AsArray.Childs.Where((x) => { return x["id"].Value == "cuabLinkTo"; }).FirstOrDefault();
			if (jnCuabLinkTo == null)
			{
				if (presetPackage != null)
				{
					presetPackage.OneNotFound();
				}
				return;
			}

			string atomType = cuaNode["atoms"].AsArray.Childs.FirstOrDefault()["type"].Value;

			string linkRBName = jnCuabLinkTo["linkTo"];
			Vector3 localPos = new Vector3(
					float.Parse(jnCuabLinkTo["posx"].Value, System.Globalization.CultureInfo.InvariantCulture),
					float.Parse(jnCuabLinkTo["posy"].Value, System.Globalization.CultureInfo.InvariantCulture),
					float.Parse(jnCuabLinkTo["posz"].Value, System.Globalization.CultureInfo.InvariantCulture));
					
			Vector3 localRot = new Vector3(
					float.Parse(jnCuabLinkTo["rotx"].Value, System.Globalization.CultureInfo.InvariantCulture),
					float.Parse(jnCuabLinkTo["roty"].Value, System.Globalization.CultureInfo.InvariantCulture),
					float.Parse(jnCuabLinkTo["rotz"].Value, System.Globalization.CultureInfo.InvariantCulture));
					
			// Log(string.Format("localPos: {0:0.00000}, {1:0.00000}, {2:0.00000}", localPos.x, localPos.y, localPos.z));
			// Log(string.Format("localRot: {0:0.00000}, {1:0.00000}, {2:0.00000}", localRot.x, localRot.y, localRot.z));
					
			CreateLinkedAtom(atomType, rootAtom, linkRBName, atomId, localPos, localRot, presetPackage: presetPackage, preset: cuaNode, atomPostfixIfExist: rootAtom.uid, onFinishCallback: (apAtom) => {
				
				linkedCuaSelector.choices = Helpers.GetLinkedAtomUids(rootAtom, CheckIncludeAtomType);
				linkedCuaSelector.val = apAtom.uid;
				updateOffsetsFromSelectedAtomQueued = true;
			});
		}

		private void CreateLinkedAtom(string atomType, Atom parentAtom, string linkRBName, string atomPrefix, Vector3 localPos, Vector3 localRot, PresetCuaPackage presetPackage = null, JSONNode preset = null, string atomPostfixIfExist = "", Action<Atom> onFinishCallback = null)
		{
			loading.LoadingStart("Creating atom...");
			foreach (Rigidbody rb in parentAtom.linkableRigidbodies)
			{
				if (rb.name == linkRBName)
				{
					atomizer.CreateAtom(atomType, atomPrefix, atomPostfixIfExist, (apAtom) =>
					{
						if (presetPackage != null)
						{
							presetPackage.Add(atomPrefix, apAtom, preset, (postPreset) => 
							{
								lastAddedAssets.Add(parentAtom, apAtom);
								UpdateAtomToLink(apAtom, parentAtom, rb, localPos, localRot, postPreset);
								if (onFinishCallback != null)
								{
									onFinishCallback(apAtom);
								}
							});
						}
						else
                        {
							lastAddedAssets.Add(parentAtom, apAtom);
							UpdateAtomToLink(apAtom, parentAtom, rb, localPos, localRot, preset);
							if (onFinishCallback != null)
							{
								onFinishCallback(apAtom);
							}
						}

						loading.LoadingEnd();
						
					});

					return;
				}
			}

			if (presetPackage != null)
			{
				presetPackage.OneNotFound();
			}

			loading.LoadingEnd();
		}

		private void RemoveLastAddedItems(Atom parentAtom)
        {
			// Remove last added clothing items
			DAZCharacterSelector geometry = Helpers.GetDAZCharacterSelector(parentAtom);
			if (geometry != null)
            {
				List<DAZClothingItem> clothingItems = lastAddedAssets.GetClothing(parentAtom);
				foreach(DAZClothingItem clothingItem in clothingItems)
                {
					geometry.SetActiveClothingItem(clothingItem, false);
				}
			}

			// Remove last added assets
			RemoveLinkedAssets(parentAtom, true);
		}

		private void RemoveLinkedAssets(Atom parentAtom, bool limitToLastAdded = false) 
		{
			if (parentAtom == null || parentAtom.destroyed) 
			{
				Log("Select a person first.");
				return;
			}

			bool forceRemoval = (removeAllLinkedAssetQuickCounter >= FORCE_REMOVE_CLICKS);

			List<Atom> atomsToRemove = null;
			foreach(Atom atom in Helpers.GetChildAtoms(parentAtom, CheckIncludeAtomType)) 
			{
				if (!forceRemoval && 
					((options.ignoreAllSceneAssets.val || options.ignoreSceneAssetsContainingScripts.val || options.ignoreSceneAssetsMarkedHidden.val) && (atomTracker.sceneCuas.Contains(atom))))
                {
					// Check if this atom should be ignored!
					if (options.ignoreAllSceneAssets.val)
                    {
						continue;
                    }
					
					if (options.ignoreSceneAssetsContainingScripts.val)
                    {
						List<Pair> scripts = Helpers.GetAtomScripts(atom);
						if (scripts != null && scripts.Count > 0)
                        {
							Log("Skipping atom " + atom.uid + " containing scripts: " + string.Join(", ", scripts.Select((x) => { return x.key + ":" + x.value; }).ToArray()));
							continue;
                        }
                    }

					if (options.ignoreSceneAssetsMarkedHidden.val && atom.hidden)
                    {
						continue;
					}

                } 

				if (limitToLastAdded && !lastAddedAssets.Contains(parentAtom, atom))
                {
					continue;
                }
				
				if (atomsToRemove == null)
				{
					atomsToRemove = new List<Atom>();
				}
				atomsToRemove.Add(atom);
			}
			if (atomsToRemove != null) 
			{
				foreach(Atom atom in atomsToRemove) 
				{
					lastAddedAssets.Remove(parentAtom, atom);
					try
					{
						atom.Remove();
					}
					catch (Exception e)
					{
						SuperController.LogError("Removing atom failed with:");
						SuperController.LogError(e.ToString());
					}
				}
			}

			if (forceRemoval)
            {
				SetRemoveAllLinkedAssetQuickCounter(0);
            }

			linkedCuaSelector.choices = Helpers.GetLinkedAtomUids(rootAtom, CheckIncludeAtomType);
			if (linkedCuaSelector.val != "" && !linkedCuaSelector.choices.Contains(linkedCuaSelector.val))
            {
				linkedCuaSelector.val = "";
			}

			// updates save preset button
			SavePresetNameCallback(savePresetName.val);
		}

		public void SaveClothingPreset()
		{
			if (rootAtom == null || rootAtom.destroyed)
			{
				Log("Select a person to base the preset on.");
				return;
			}

			PresetFileInfo presetInfo = pluginPath.GetSavePresetFileInfo(savePresetName.val);
			if (!presetInfo.isValid)
			{
				Log("No valid preset name selected.");
				return;
			}

			List<DAZClothingItem> activeClothing = Helpers.GetClothingItems(rootAtom).ToList();
			if (activeClothing == null || activeClothing.Count == 0)
            {
				Log("Selected person has no active clothing items.");
				return;
			}

			System.Text.StringBuilder sb = new System.Text.StringBuilder();
			bool firstCreated = false;
			sb.AppendLine("{");
			sb.AppendLine("  \"id\" : \"" + presetInfo.presetName + "\",");
			sb.AppendLine("  \"type\" : \"" + TYPE_CLOTHING + "\",");
			sb.AppendLine(" \"clothing\": [");
			foreach(DAZClothingItem clothing in activeClothing)
            {
				if (firstCreated)
				{
					sb.AppendLine(",");
				}
				JSONClass jc = Helpers.CreateJSONClothingClass(clothing);
				sb.AppendLine(jc.ToString());
				firstCreated = true;
			}
			sb.AppendLine("]");
			sb.AppendLine("}");

			JSONClass mainjc = JSONNode.Parse(sb.ToString()).AsObject;
			SuperController.singleton.SaveJSON(mainjc, presetInfo.filePath);

			Log("Preset '" + presetInfo.presetName + "' saved");

			// updates save preset button
			SavePresetNameCallback(savePresetName.val);
			loadPresetNameSelector.choices = pluginPath.GetLoadPresetSelectionNames(true);
			// Select newly created preset
			string selectablePresetName = loadPresetNameSelector.choices.FirstOrDefault((x) => { return x == presetInfo.presetName; });
			loadPresetNameSelector.val = (selectablePresetName == null ? "" : selectablePresetName);
			options.SuggestLoadPreset(loadPresetNameSelector.val);
		}

		private void SavePreset()
		{
			if (rootAtom == null || rootAtom.destroyed) {
				Log("Select a person to base the preset on.");
				return;
			}
			
			PresetFileInfo presetInfo = pluginPath.GetSavePresetFileInfo(savePresetName.val);
			if (!presetInfo.isValid) {
				Log("No valid preset name selected.");
				return;
			}

			List<Atom> childAtoms = Helpers.GetChildAtoms(rootAtom, CheckIncludeAtomType).ToList();
			if (childAtoms.Count == 0)
            {
				Log("No linked assets for selected person. Nothing to save.");
				return;
			}

			System.Text.StringBuilder sb = new System.Text.StringBuilder();
			bool firstCreated = false;
			sb.AppendLine("{");
			sb.AppendLine("  \"id\" : \"" + presetInfo.presetName + "\",");
			sb.AppendLine("  \"type\" : \"" + TYPE_LINKED_ASSET + "\",");
			sb.AppendLine(" \"items\": [");
			foreach(Atom atom in childAtoms) 
			{
				if (firstCreated) {
					sb.AppendLine(",");
				}
				JSONClass jc = Helpers.CreateJSONClass(rootAtom, atom);
				sb.AppendLine(jc.ToString());
				firstCreated = true;
			}
			sb.AppendLine("]");
			sb.AppendLine("}");
			JSONClass mainjc = JSONNode.Parse(sb.ToString()).AsObject;
			SuperController.singleton.SaveJSON(mainjc, presetInfo.filePath);
			
			Log("Preset '" + presetInfo.presetName + "' saved");
			
			// updates save preset button
			SavePresetNameCallback(savePresetName.val);
			loadPresetNameSelector.choices = pluginPath.GetLoadPresetSelectionNames(true);
			// Select newly created preset
		 	string selectablePresetName = loadPresetNameSelector.choices.FirstOrDefault((x) => { return x == presetInfo.presetName; });
			loadPresetNameSelector.val = (selectablePresetName == null ? "" : selectablePresetName);
			options.SuggestLoadPreset(loadPresetNameSelector.val);
		}
	
		
		private void SelectRootAtom(string uid)
		{
		 	Atom rAtom = SuperController.singleton.GetAtomByUid(uid);
			
			if (rAtom == null || rAtom.destroyed) 
			{
				rAtom = null;
			}
			// Force root atom to be visible in the selector if not already selected
			if (rAtom == null || rootAtomSelector.choices == null || !rootAtomSelector.choices.Contains(rAtom.uid)) {
				List<string> forcedSelection = new List<string>();
				if (rAtom != null)
				{
					forcedSelection.Add(rAtom.uid);
				}
				rootAtomSelector.choices = forcedSelection;
				rootAtomSelector.valNoCallback = rAtom.uid;
			} else if (rootAtomSelector.val != rAtom.uid) {
				rootAtomSelector.valNoCallback = rAtom.uid;
			}
			
			rootAtom = rAtom;
			selectedLinkedAtom = null;
			// deselect save-preset
			savePresetName.val = "";
			savePresetNameSuggestionSelector.val = "";
			linkedCuaSelector.choices = Helpers.GetLinkedAtomUids(rootAtom, CheckIncludeAtomType);
			linkedCuaSelector.val = (linkedCuaSelector.choices.Count > 0 ? linkedCuaSelector.choices[0] : "");
			updateOffsetsFromSelectedAtomQueued = true;
		}
		
		private void CreateLinkableAssets(params LinkableAsset[] assets) 
		{
			int rightSpaces = 0;
			Dictionary<string, int> prefixCounts = new Dictionary<string, int>();
			foreach(LinkableAsset asset in assets) 
			{
				if (!asset.guiRightSide) 
				{
					rightSpaces += 1;
					if (rightSpaces > 1) 
					{
						CreateSpacer(true).height = 50;
					}
				} 
				else 
				{
					rightSpaces = 0;
				}
				
				int prefixNameCount = 0;
				if (!prefixCounts.TryGetValue(asset.atomPrefix, out prefixNameCount)) 
				{
					prefixNameCount = 0;
				}
				if (prefixNameCount > 0) 
				{
					asset.atomPrefix += "_" + prefixNameCount;
				}
				
				UIDynamicButton btn = uiHelper.CreateButton(asset.name, asset.guiRightSide);
				btn.buttonColor = new Color(0.9f, 0.9f, 0.9f);
				btn.button.onClick.AddListener(() =>
				{
					try {
						if (loading.LoadingInProgress())
						{
							return;
						}
						string atomType = createGenericAssetTypeSelector.val;
						if (atomType != null && atomType != "")
						{
							if (selectedBtn != null) 
							{
								selectedBtn.buttonColor = new Color(0.9f, 0.9f, 0.9f);
							}
							selectedBtn = btn;
							selectedBtn.buttonColor = new Color(1f, 0.8f, 0.8f);
							lastAddedAssets.Clear(rootAtom);

							CreateLinkedAtom(atomType, rootAtom, asset.linkRBName, asset.atomPrefix, asset.localPosition, asset.localRotation, onFinishCallback: (apAtom) =>
							{
								// Initial pos/rot state
								apAtom.mainController.currentPositionState = FreeControllerV3.PositionState.ParentLink;
								apAtom.mainController.currentRotationState = FreeControllerV3.RotationState.ParentLink;

								// disable collision as default for new atoms
								apAtom.collisionEnabledJSON.val = false;
								apAtom.GetStorableByID("control")?.SetBoolParamValue("physicsEnabled", false);
								
								// update linkedCuaSelector after creating atom
								linkedCuaSelector.choices = Helpers.GetLinkedAtomUids(rootAtom, CheckIncludeAtomType);
								linkedCuaSelector.val = apAtom.uid;
								updateOffsetsFromSelectedAtomQueued = true;
							});
						}
					} catch (Exception e) {
						SuperController.LogError(e.ToString());
					}
				});
				
				prefixNameCount += 1;
				prefixCounts[asset.atomPrefix] = prefixNameCount;
			}
		}

		/// <summary>
        /// Updates the plugin offset pos/rot values and other values in the plugin GUI from selected atom/link.
        /// </summary>
		private void UpdateOffsetsFromAtom(Atom apAtom, Atom parentAtom, Rigidbody rbLink)
		{
			float s = 1f;
			posRotSliders.UpdateOffsetsPosRotFromAtom(apAtom, parentAtom, rbLink);

			JSONStorable jsonScale = apAtom.GetStorableByID("scale");
			s = (jsonScale != null && jsonScale.IsFloatJSONParam("scale") ? jsonScale.GetFloatParamValue("scale") : 1f);

			scaleModifier.valNoCallback = s;
            canUpdatePosition.valNoCallback = apAtom.mainController.canGrabPosition;
            canUpdateRotation.valNoCallback = apAtom.mainController.canGrabRotation;
			atomOn.valNoCallback = apAtom.on;
			atomHidden.valNoCallback = apAtom.hidden;
			atomLinkTypeSelector.valNoCallback = options.GetAtomLinkType(apAtom);

			atomCollision.valNoCallback = apAtom.collisionEnabledJSON.val;
			atomPhysicsEnabled.valNoCallback = apAtom.GetStorableByID("control").GetBoolParamValue("physicsEnabled");
			atomUseGravity.valNoCallback = apAtom.GetStorableByID("control").GetBoolParamValue("useGravity");

			List<Pair> scripts = Helpers.GetAtomScripts(apAtom);
			bool containsScripts = (scripts != null && scripts.Count > 0);

			if (apAtom.type == "CustomUnityAsset")
			{
				AssetDataInfo data = Helpers.GetAtomAssetInfo(apAtom);
				selectedAtomInfo.val = "bundle:\n" + data.assetBundleName + "\n\nasset:\n" + data.assetName + "\n\nurl:\n" + data.assetUrl + "\n\n" + (containsScripts ? "Atom contain scripts: yes\n" : "");
			} 
			else
            {
				selectedAtomInfo.val = "Atom type: " + apAtom.type;
			}
		}

		private void UpdateAtomToLink(Atom apAtom, Atom parentAtom, Rigidbody linkRB, Vector3 localPos, Vector3 localRot, JSONNode preset = null, bool applyPluginValues = false)
		{
			if (!inUpdate)
            {
				Log("UpdateAtomToLink() is not running in main update! This could cause unknown problems/racing conditions.");
			}
			if (applyPluginValues)
			{
				localPos += posRotSliders.Position;
				localRot += posRotSliders.Rotation;
			}

			Vector3 pos = linkRB.transform.TransformPoint(localPos);
			Quaternion rot = (linkRB.transform.rotation * Quaternion.Inverse(Quaternion.Euler(localRot)));

			if (applyPluginValues)
			{
				apAtom.mainController.canGrabPosition = canUpdatePosition.val;
				apAtom.mainController.canGrabRotation = canUpdateRotation.val;
				apAtom.hidden = atomHidden.val;
				apAtom.SetOn(atomOn.val);
				apAtom.collisionEnabledJSON.val = atomCollision.val;
				apAtom.GetStorableByID("control").SetBoolParamValue("physicsEnabled", atomPhysicsEnabled.val);
				apAtom.GetStorableByID("control").SetBoolParamValue("useGravity", atomUseGravity.val);

				FreeControllerV3.PositionState targetPosState = options.GetAtomLinkTypePosition(atomLinkTypeSelector.val, apAtom.mainController.currentPositionState);
				FreeControllerV3.RotationState targetRotState = options.GetAtomLinkTypeRotation(atomLinkTypeSelector.val, apAtom.mainController.currentRotationState);

				// When changing LinkType also keep position/rotation intact (don't change it).
				if (apAtom.mainController.currentPositionState != targetPosState)
				{
					apAtom.mainController.currentPositionState = targetPosState;
					pos = (apAtom.childAtomContainer != null ? apAtom.childAtomContainer.transform.position : apAtom.transform.position);
					updateOffsetsFromSelectedAtomQueued = true;
				}
				if (apAtom.mainController.currentRotationState != targetRotState)
				{
					apAtom.mainController.currentRotationState = targetRotState;
					rot = (apAtom.childAtomContainer != null ? apAtom.childAtomContainer.transform.rotation : apAtom.transform.rotation);
					updateOffsetsFromSelectedAtomQueued = true;
				}
			}

			if (!updateOffsetsFromSelectedAtomQueued && preset == null)
			{
				if (options.TrackLinkInWorldSpace(apAtom.mainController.currentPositionState))
				{
					pos = localPos;
				}
				if (options.TrackLinkInWorldSpace(apAtom.mainController.currentRotationState))
				{
					rot = Quaternion.Euler(localRot);
				}
			}

			// Unlink in case user updates the atom position (can't move around 100% of the time if it's not unlinked...?)
			apAtom.mainController.SetLinkToAtom("None");
			apAtom.mainController.linkToRB = null;
			
			apAtom.mainController.transform.SetPositionAndRotation(pos, rot);
			apAtom.childAtomContainer?.transform?.SetPositionAndRotation(pos, rot);

			FreeControllerV3.PositionState presetPosState = FreeControllerV3.PositionState.ParentLink;
			FreeControllerV3.RotationState presetRotState = FreeControllerV3.RotationState.ParentLink;

			if (preset != null) 
			{
				JSONArray asArray = preset["atoms"].AsArray;
			    JSONClass asObject = asArray[0].AsObject;

				apAtom.PreRestore();
				// apAtom.RestoreTransform(...); <-- Never got this to work properly, but not required as transform is restored after preset is loaded down below
				apAtom.Restore(asObject, restorePhysical: true, restoreAppearance: true, restoreCore: true, presetAtoms: asArray);
				apAtom.LateRestore(asObject, restorePhysical: true, restoreAppearance: true, restoreCore: true);
				apAtom.PostRestore();

				// Avoid PhysicsLink and PhysicsEnabled directly after creation and postpone setting these values when a preset is used.
				if (options.delayPhysicsProperties.val && 
					(apAtom.mainController.currentPositionState == FreeControllerV3.PositionState.PhysicsLink || 
					 apAtom.mainController.currentRotationState == FreeControllerV3.RotationState.PhysicsLink ||
					 apAtom.collisionEnabledJSON.val ||
					 apAtom.GetStorableByID("control").GetBoolParamValue("physicsEnabled")))
				{
					UpdateAtomPostCreation postArgs = new UpdateAtomPostCreation();
					postArgs.apAtom = apAtom;
					postArgs.posState = apAtom.mainController.currentPositionState;
					postArgs.rotState = apAtom.mainController.currentRotationState;
					postArgs.collisionEnabled = apAtom.collisionEnabledJSON.val;
					postArgs.physicsEnabled = apAtom.GetStorableByID("control").GetBoolParamValue("physicsEnabled");
					postArgs.delayFrames = 50;
					updateAtomPostCreationQueue.Enqueue(postArgs);

					if (apAtom.mainController.currentPositionState == FreeControllerV3.PositionState.PhysicsLink)
						apAtom.mainController.currentPositionState = FreeControllerV3.PositionState.ParentLink;
					if (apAtom.mainController.currentRotationState == FreeControllerV3.RotationState.PhysicsLink)
						apAtom.mainController.currentRotationState = FreeControllerV3.RotationState.ParentLink;
					apAtom.collisionEnabledJSON.val = false;
					apAtom.GetStorableByID("control").SetBoolParamValue("physicsEnabled", false);
				}

				presetPosState = apAtom.mainController.currentPositionState;
				presetRotState = apAtom.mainController.currentRotationState;

			}

			// Unlink in case user updates the atom position (can't move around 100% of the time if it's not unlinked...?)
			apAtom.mainController.SetLinkToAtom("None");
			apAtom.mainController.linkToRB = null;
			
			apAtom.mainController.transform.SetPositionAndRotation(pos, rot);
			apAtom.childAtomContainer?.transform?.SetPositionAndRotation(pos, rot);

			if (applyPluginValues && options.enableScalingModifier.val)
			{
				JSONStorable jsonScale = apAtom.GetStorableByID("scale");
				if (jsonScale != null)
                {
					jsonScale.SetFloatParamValue("scale", scaleModifier.val);
				}
			}

			// Link atom
			apAtom.mainController.SetLinkToAtom(parentAtom.uid);
			apAtom.mainController.linkToRB = linkRB;

			if (preset != null)
            {
				apAtom.mainController.currentPositionState = presetPosState;
				apAtom.mainController.currentRotationState = presetRotState;
			} 
			else if (applyPluginValues)
            {
				Notifier.NotifyAtomUpdatedToPlugins(apAtom);
			}
		}

		private void Log(string msg)
		{
			SuperController.LogMessage(msg);
		}
	}
}