using System;
using System.Text;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Blazedust.CUAM
{
	public class Options
	{
		CustomUnityAssetManager core;

		public JSONStorableBool autoSelectPersonFromScene;
		public JSONStorableBool autoSelectAssetController;
		public JSONStorableBool delayPhysicsProperties;
		public JSONStorableBool enableScalingModifier;
		public JSONStorableBool ignoreAllSceneAssets;
		public JSONStorableBool ignoreSceneAssetsContainingScripts;
		public JSONStorableBool ignoreSceneAssetsMarkedHidden;
		public JSONStorableBool smartRemovalOfPresets;
		public JSONStorableBool enableLoadTriggers;
		public JSONStorableBool updateScriptAtomUids;
		public JSONStorableBool enablePrioritizedLoadingTriggers;
		public JSONStorableBool enablePlayModeUpdates;
		public JSONStorableBool enablePersonUidLoadTriggers;
		public JSONStorableString autoLoaderPresetsList;
		public JSONStorableString includedAtomTypesList;
		public JSONStorableStringChooser addLoadTriggerPresetSelection;
		public JSONStorableStringChooser addLoadTriggerPersonUidSelection;
		public JSONStorableStringChooser addLoadTriggerMorphSelection;

		private List<AutoLoadTriggerInfo> autoLoadTriggerPresets;
		private bool resetActiveMorphsAfterLoading;
		private int autoLoadCheckerCounter;
		private Dictionary<Atom, AtomTriggerStates> activeTriggersPerAtom;
		private List<string> includedAtomTypes;

		public const string ANY_MORPH = "* ANY *";

		private Atom lastRootAtom;
		private UIHelper uiHelper;

		public Options(CustomUnityAssetManager c)
        {
			core = c;
			uiHelper = new UIHelper(c);

			resetActiveMorphsAfterLoading = true;
			autoLoadTriggerPresets = new List<AutoLoadTriggerInfo>();
			activeTriggersPerAtom = new Dictionary<Atom, AtomTriggerStates>();
			autoLoadCheckerCounter = 0;
			lastRootAtom = null;
			includedAtomTypes = new List<string>();

			autoSelectPersonFromScene = new JSONStorableBool("cuabAutoSelectPerson", true, SelectedOptionsToggleChanged);
			autoSelectAssetController = new JSONStorableBool("cuabOptionAutoSelectAssetController", true, SelectedOptionsToggleChanged);
			delayPhysicsProperties = new JSONStorableBool("cuabOptionDelayPhysicsProperties", true, SelectedOptionsToggleChanged);
			enableScalingModifier = new JSONStorableBool("cuabOptionEnableScaling", true, SelectedOptionsToggleChanged);
			enableLoadTriggers = new JSONStorableBool("cuabOptionEnableLoadTriggers", true, SelectedOptionsToggleChanged);
			ignoreAllSceneAssets = new JSONStorableBool("cuabOptionIgnoreAllSceneAssets", true, SelectedOptionsToggleChanged);
			ignoreSceneAssetsContainingScripts = new JSONStorableBool("cuabOptionIgnoreSceneAssetsContainingScripts", true, SelectedOptionsToggleChanged);
			ignoreSceneAssetsMarkedHidden = new JSONStorableBool("cuabOptionIgnoreSceneAssetsMarkedHidden", true, SelectedOptionsToggleChanged);
			smartRemovalOfPresets = new JSONStorableBool("cuabOptionSmartRemovalOfPresets", true, SelectedOptionsToggleChanged);
			enablePersonUidLoadTriggers = new JSONStorableBool("cuabOptionEnablePersonUidLoadTriggers", true, SelectedOptionsToggleChanged);
			autoLoaderPresetsList = new JSONStorableString("cuabAutoLoaderPresetsList", "");
			includedAtomTypesList = new JSONStorableString("cuabIncludedAtomTypesList", "");
			updateScriptAtomUids = new JSONStorableBool("cuabUpdateScriptAtomUids", true, SelectedOptionsToggleChanged);
			enablePrioritizedLoadingTriggers = new JSONStorableBool("cuabEnablePrioritizedLoadingTriggers", false, SelectedOptionsToggleChanged);
			enablePlayModeUpdates = new JSONStorableBool("cuabEnablePlayModeUpdates", false, SelectedOptionsToggleChanged);
		}

		public void Destroy()
		{
			uiHelper.Destroy();
			uiHelper = null;
			lastRootAtom = null;
			core = null;
			activeTriggersPerAtom.Clear();
			activeTriggersPerAtom = null;
			autoLoadTriggerPresets.Clear();
			autoLoadTriggerPresets = null;
		}

		public void InitGui()
        {
			uiHelper.CreateSpacer(true, 50);

			UIDynamicButton btn = uiHelper.CreateButton("Plugin options: ", false);
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					SuperController.singleton.ClearMessages();
					SuperController.LogMessage("");
					SuperController.LogMessage("...Scroll down to read everything...");
					SuperController.LogMessage("");
					SuperController.LogMessage("Options");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Auto Select Person From Scene - Keeps the selected person in the scene as the active person in the plugin.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Auto Select Asset Controller - Selects the selected asset controller in the scene automatically so you can grab and move/rotate the controller in the scene without leaving the plugin view.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Delay Physic Properties On Load Preset - Delay setting PhysicLink, PhysicsEnabled and Collision properties on created atoms to prevent crazy physics.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Enable Scaling Modifier - Disable this if you have other plugins that might handle scaling on linked atoms instead.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Update AtomUids In Scripts And Properties - Update any referenced atomUids in scripts and properties to match the created atom's uid in case of duplicate uids.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Enable Prioritized Loading Triggers - If you have two or more triggers using the same trigger morph, personUid will be used to narrow down the hit result to the most prioritized trigger. This allows you to have variations of the same look with different personUid.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Enable Update In Play Mode - Normally you change look and clothing for a person while in Edit Mode. This allows the plugin to make a minimal impact while in Play Mode. If you need this plugin to also check for look changes and clothing changes while in Play Mode then you should enable this option.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Filter: Ignore All Scene Assets - Scene assets will be ignored when 'Loading Presets' and 'Remove All Linked Assets' are used. Click 'Remove All Linked Assets' several times to force removal and ignore this rule.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Filter: Ignore Scene Assets Containing Scripts - Same as Ignore Scene Assets but the asset will be ignored if it contains any script.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Filter: Ignore Scene Assets Marked Hidden - Same as Ignore Scene Assets but the asset will be ignored if it is hidden.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Enable Load Triggers - If load triggers should be enabled. Load triggers automatically load presets based on morph and/or personUid on a character. Here you set up that morph X and personUid Y should load preset Z, it's all managed with the options to the right. Make sure to use a strong unique person uid if you choose to also use this option.");
					SuperController.LogMessage(" NOTE: Strongly recommended to ONLY trigger on morphs as changing looks in a scene can update the uid with a postfix if it's not unique and any load triggers operating on personUids won't work! This plugin comes bundled with 15 empty morphs you can save togheter with your look to have a unique morph to trigger on.");
					SuperController.LogMessage(" NOTE: Adding an auto load trigger for the first time will automatically trigger it at once and load the selected preset.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Enable Smart Auto-Load-Trigger Preset Removal - Analyzes scene atoms linked to a person to determine if they belong to an existing preset and should be unloaded when a look is changed. This will allow the plugin to properly remove assets belonging to a look but prevent removing linked assets that belongs to the scene. Requires 'Enable Load Triggers' to be enabled.");
					SuperController.LogMessage("");
					SuperController.LogMessage("* Atom Types To Handle - You can add additional atom types here that you want the plugin to handle.");
					SuperController.LogMessage("");
					SuperController.LogMessage("...Scroll to top to read everything...");
					SuperController.singleton.OpenMessageLogPanel();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			UIDynamicToggle toggle = uiHelper.CreateToggle(autoSelectPersonFromScene, false);
			toggle.labelText.text = "Auto Select Person From Scene";

			toggle = uiHelper.CreateToggle(autoSelectAssetController, false);
			toggle.label = "Auto Select Asset Controller";
			
			toggle = uiHelper.CreateToggle(delayPhysicsProperties, false);
			toggle.height = 80;
			toggle.label = "Delay Physic Properties On Load Preset";

			toggle = uiHelper.CreateToggle(enableScalingModifier, false);
			toggle.label = "Enable Scaling Modifier";

			toggle = uiHelper.CreateToggle(updateScriptAtomUids, false);
			toggle.height = 80;
			toggle.label = "Update AtomUids In Scripts And Properties";

			toggle = uiHelper.CreateToggle(enablePrioritizedLoadingTriggers, false);
			toggle.height = 40;
			toggle.label = "Enable Prioritized Loading Triggers";

			toggle = uiHelper.CreateToggle(enablePlayModeUpdates, false);
			toggle.height = 40;
			toggle.label = "Enable Update In Play Mode";

			toggle = uiHelper.CreateToggle(ignoreAllSceneAssets, true);
			toggle.height = 80;
			toggle.label = "Filter for Replace/Remove:\nIgnore All Scene Assets";
			toggle.backgroundColor = new Color(0.9f, 0.8f, 0.8f);

			toggle = uiHelper.CreateToggle(ignoreSceneAssetsContainingScripts, true);
			toggle.height = 120;
			toggle.label = "Filter for Replace/Remove:\nIgnore Scene Assets Containing Scripts";
			toggle.backgroundColor = new Color(0.9f, 0.8f, 0.8f);

			toggle = uiHelper.CreateToggle(ignoreSceneAssetsMarkedHidden, true);
			toggle.height = 120;
			toggle.label = "Filter for Replace/Remove:\nIgnore Scene Assets Marked Hidden";
			toggle.backgroundColor = new Color(0.9f, 0.8f, 0.8f);

			uiHelper.CreateSpacer(false, 40);
			uiHelper.CreateSpacer(true, 40);

			toggle = uiHelper.CreateToggle(smartRemovalOfPresets, true);
			toggle.height = 80;
			toggle.label = "Enable Smart Auto-Load-Trigger Preset Removal";
			toggle.backgroundColor = new Color(0.8f, 0.9f, 0.9f);

			toggle = uiHelper.CreateToggle(enableLoadTriggers, true);
			toggle.label = "Enable Load Triggers";
			toggle.backgroundColor = new Color(0.8f, 0.9f, 0.9f);

			uiHelper.CreateSpacer(true, 80);

			uiHelper.CreateLabel("Auto load preset triggers", true);

			addLoadTriggerMorphSelection = new JSONStorableStringChooser("cuabAddTriggerMorphSelection", GetMorphTriggerNames(), "", "Trigger Morph", (string triggerMorphName) => { });
			UIDynamicPopup popup = uiHelper.CreateScrollablePopup(addLoadTriggerMorphSelection, true);
			popup.popupPanelHeight = 700f;
			addLoadTriggerMorphSelection.popup.onOpenPopupHandlers += () => {
				addLoadTriggerMorphSelection.choices = GetMorphTriggerNames();
			};

			addLoadTriggerPersonUidSelection = new JSONStorableStringChooser("cuabAddTriggerPersonUidSelection", GetPersonUidTriggerNames(), "", "(Optional) Trigger Person Uid", (string personUid) => { });
			popup = uiHelper.CreateScrollablePopup(addLoadTriggerPersonUidSelection, true);
			popup.popupPanelHeight = 700f;
			addLoadTriggerPersonUidSelection.popup.onOpenPopupHandlers += () => {
				addLoadTriggerPersonUidSelection.choices = GetPersonUidTriggerNames();
			};

			addLoadTriggerPresetSelection = new JSONStorableStringChooser("cuabAddTriggerPresetSelection", core.pluginPath.GetLoadPresetSelectionNames(), "", "Load Preset");
			popup = uiHelper.CreateScrollablePopup(addLoadTriggerPresetSelection, true);
			popup.popupPanelHeight = 700f;
			addLoadTriggerPresetSelection.popup.onOpenPopupHandlers += () => {
				addLoadTriggerPresetSelection.choices = core.pluginPath.GetLoadPresetSelectionNames();
			};

			btn = uiHelper.CreateButton("Add/Create Trigger", true);
			btn.buttonColor = new Color(0.7f, 1.0f, 0.7f);
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					if (addLoadTriggerMorphSelection.val != "" && addLoadTriggerPresetSelection.val != "")
					{
						AddLoadTrigger(addLoadTriggerPersonUidSelection.val, addLoadTriggerMorphSelection.val, addLoadTriggerPresetSelection.val);
						addLoadTriggerPersonUidSelection.val = "";
						addLoadTriggerMorphSelection.valNoCallback = "";
						addLoadTriggerPresetSelection.valNoCallback = "";
					} 
					else
                    {
						SuperController.LogMessage("Auto Triggers requires a Trigger Morph to be selected!");
                    }
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});


			JSONStorableStringChooser removeLoadTriggersSelection = new JSONStorableStringChooser("cuabRemoveTriggerSelection", GetLoadTriggerNames(), "", "Selected Trigger");
			popup = uiHelper.CreateScrollablePopup(removeLoadTriggersSelection, true);
			popup.popupPanelHeight = 700f;
			removeLoadTriggersSelection.popup.onOpenPopupHandlers += () => {
				removeLoadTriggersSelection.choices = GetLoadTriggerNames();
			};

			btn = uiHelper.CreateButton("Remove Selected Trigger", true);
			btn.buttonColor = new Color(1.0f, 0.7f, 0.7f);
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					if (removeLoadTriggersSelection.val != "")
					{
						RemoveLoadTriggerName(removeLoadTriggersSelection.val);
						removeLoadTriggersSelection.valNoCallback = "";
					}
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			UIDynamicTextField text = uiHelper.CreateTextField(autoLoaderPresetsList, true);
			text.height = 700;

			uiHelper.CreateSpacer(false, 85);
			uiHelper.CreateLabel("Atom types to handle (assets):", false);

			// 
			List<string> availableAtomTypes = new List<string>();
			// availableAtomTypes.Add("CustomUnityAsset"); <-- Forced included, always
			availableAtomTypes.Add("Light/InvisibleLight");
			availableAtomTypes.Add("Light/Torch");

			availableAtomTypes.Add("Misc/Empty");
			availableAtomTypes.Add("Misc/ImagePanel");
			availableAtomTypes.Add("Misc/ImagePanelEmissive");
			availableAtomTypes.Add("Misc/SimpleSign");
			availableAtomTypes.Add("Misc/UIText");
			// Browser might not make too much sense for the purpose of this plugin but add anyway...
			availableAtomTypes.Add("Misc/WebBrowser");
			availableAtomTypes.Add("Misc/WebPanel");
			availableAtomTypes.Add("Misc/WebPanelEmissive");

			availableAtomTypes.Add("Toys/Dildo");
			availableAtomTypes.Add("Toys/Paddle");
			availableAtomTypes.Add("Toys/ToyAH");
			availableAtomTypes.Add("Toys/ToyBP");

			availableAtomTypes.Add("Shapes/Capsule");
			availableAtomTypes.Add("Shapes/Cube");
			availableAtomTypes.Add("Shapes/ISCapsule");
			availableAtomTypes.Add("Shapes/ISCone");
			availableAtomTypes.Add("Shapes/ISCube");
			availableAtomTypes.Add("Shapes/ISCylinder");
			availableAtomTypes.Add("Shapes/ISSphere");
			availableAtomTypes.Add("Shapes/ISTube");
			availableAtomTypes.Add("Shapes/Sphere");

			// Browser might not make too much sense for the purpose of this plugin but add anyway...
			availableAtomTypes.Add("Triggers/Button");
			availableAtomTypes.Add("Triggers/CollisionTrigger");
			availableAtomTypes.Add("Triggers/LookAtTrigger");
			availableAtomTypes.Add("Triggers/UIButton");
			availableAtomTypes.Add("Triggers/VariableTrigger");

			availableAtomTypes.Add("Force/CycleForce"); // cycle force for breathing kind of presets.
			// availableAtomTypes.Add("Force/GrabPoint"); // <-- Can't get this one to work properly. Never add to this plugin! F**ks up everything! :)
			availableAtomTypes.Add("Force/SyncForce");
			
			// availableAtomTypes.Add("People/Person"); // <-- Not supported at all, plugin uses Person as a root atom. Never add!

			JSONStorableStringChooser includedAtomTypeSelection = new JSONStorableStringChooser("cuabAddAtomTypeSelection", availableAtomTypes, "", "Select Atom Type");
			popup = uiHelper.CreateScrollablePopup(includedAtomTypeSelection, false);
			popup.popupPanelHeight = 700f;

			btn = uiHelper.CreateButton("Add Atom Type", false);
			btn.buttonColor = new Color(0.7f, 1.0f, 0.7f);
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					string atomType = includedAtomTypeSelection.val;
					if (atomType != null && atomType != "" && atomType.Contains("/"))
						atomType = atomType.Substring(atomType.IndexOf("/") + 1);

					if (atomType != null && atomType != "" && !includedAtomTypes.Contains(atomType))
					{
						includedAtomTypes.Add(atomType);
						SelectedOptionsToggleChanged(true); // trigger save
						UpdateIncludedAtomTypesList();
					}
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			btn = uiHelper.CreateButton("Remove Atom Type", false);
			btn.buttonColor = new Color(1.0f, 0.7f, 0.7f);
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					string atomType = includedAtomTypeSelection.val;
					if (atomType != null && atomType != "" && atomType.Contains("/"))
						atomType = atomType.Substring(atomType.IndexOf("/") + 1);
					
					if (atomType != null && atomType != "" && includedAtomTypes.Contains(atomType))
					{
						includedAtomTypes.Remove(atomType);
						SelectedOptionsToggleChanged(true); // trigger save
						UpdateIncludedAtomTypesList();
					}
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			text = uiHelper.CreateTextField(includedAtomTypesList, false);
			text.height = 600;

			btn = uiHelper.CreateButton("Plugin quirks (?)", true);
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					SuperController.singleton.ClearMessages();
					SuperController.LogMessage("* QUIRKS:");
					SuperController.LogMessage("");
					SuperController.LogMessage("If you remove a CustomUnityAsset atom before the assetbundle has finish loading, VaM's CustomUnityAssetLoader will lock itself and a hard reset must be performed.");
					SuperController.LogMessage("I couldn't figure out a way to properly see if it's currently loading so this plugin will just prevent you from double-clicking buttons for a short while.");
					SuperController.LogMessage("If you intend to load assetbundles that takes a few seconds to load you must wait for the loading to complete before removing any atoms.");
					SuperController.LogMessage("");
					SuperController.LogMessage("Auto load triggers make a lot of assumtions when clothing item presets should be loaded or unloaded depending on the look/morphs.");
					SuperController.LogMessage("The plugin assumes you're doing manual edits if only one clothing item is enabled/disabled. To avoid false positives from this assumtion when loading clothing presets in certain situations you'll need to double-click when disabling clothing items bound to an auto load trigger.");
					SuperController.LogMessage("");
					SuperController.singleton.OpenMessageLogPanel();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			btn = uiHelper.CreateButton("External option files (?)", true);
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					SuperController.singleton.ClearMessages();
					SuperController.LogMessage("For creators: This plugin reads all option files and preset files in the plugin's preset folder. This allows you to share options and preset files for this plugin.");
					SuperController.LogMessage("");
					SuperController.LogMessage("All your preset files starting with 'preset_' can be shared directly.");
					SuperController.LogMessage("");
					SuperController.LogMessage("Option files are shared by creating a file named 'options_MYOPTIONS.json' in this plugin's preset folder containing the option keys 'autoMorphLoaders' and 'includedAtomTypes' which configures auto load triggers.");
					SuperController.LogMessage("");
					SuperController.LogMessage("Simply copy your options.json file and only keep the configuration keys mentioned above and rename the file to 'options_MYOPTIONS.json' where MYOPTIONS is some unique name for your file.");
					SuperController.LogMessage("");
					SuperController.LogMessage("Note: The plugin files must be in this plugin's preset folder, so it won't support .vac or .var packaging. You can always share the plugin files in a separate .zip file.");
					SuperController.singleton.OpenMessageLogPanel();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

		}

		public void SuggestLoadPreset(string loadPreset)
        {
			// if (addLoadTriggerPresetSelection.val == null || addLoadTriggerPresetSelection.val == "")
			// {
			addLoadTriggerPresetSelection.val = loadPreset;
			// }
		}
		
		public void Reset()
        {
			resetActiveMorphsAfterLoading = true;
		}

		public void Update(Atom personAtom)
        {
			if (enableLoadTriggers.val)
			{
				if (lastRootAtom != personAtom)
                {
					// When root atom changes, delay a few updates before updating auto-loaders.
					lastRootAtom = personAtom;
					autoLoadCheckerCounter = 35;
				}
				if (personAtom != null && !personAtom.destroyed && (autoLoadCheckerCounter++) % 61 == 0) // check each 61 updates, good enough!
				{
					autoLoadCheckerCounter = 1;
					CheckAutoLoaders(personAtom);
				}
			}
		}

		/// <summary>
		/// Prepare current active morphs when selecting a new person not yet selected. 
		/// This will allow CheckAutoLoaders to properly work depending on how fast the user clicks around in the scene and GUI and load other looks.
		/// </summary>
		/// <param name="atom">Person atom type</param>
		public void PersonSelected(Atom atom)
        {
			if (!atom || atom.destroyed)
				return;

			if (enableLoadTriggers.val && !activeTriggersPerAtom.ContainsKey(atom))
			{
				CheckAutoLoaders(atom);
				autoLoadCheckerCounter = 10;
			}
		}

		/// <summary>
		/// Forces a morph changed check when deselecting a person.
		/// This will allow CheckAutoLoaders to properly work depending on how fast the user clicks around in the scene and GUI and load other looks.
		/// </summary>
		/// <param name="atom"></param>
		public void PersonDeselected(Atom atom)
        {
			if (!atom || atom.destroyed)
				return;

			if (enableLoadTriggers.val && activeTriggersPerAtom.ContainsKey(atom))
			{
				CheckAutoLoaders(atom);
				autoLoadCheckerCounter = 10;
			}
		}
		
		private void CheckAutoLoaders(Atom atom)
		{
			if (atom == null || atom.destroyed)
				return;

			DAZCharacterSelector geometry = atom.GetStorableByID("geometry") as DAZCharacterSelector;
			if (geometry == null)
			{
				return;
			}

			int loadMode = 1; // 1 ALL , 2 = CLOTHING ONLY
			if (resetActiveMorphsAfterLoading)
			{
				activeTriggersPerAtom.Clear();
				resetActiveMorphsAfterLoading = false;
				// Don't load changes after initial loading
				loadMode = 2;
			}

			AtomTriggerStates state = null;
			if (!activeTriggersPerAtom.TryGetValue(atom, out state))
			{
				state = new AtomTriggerStates();
				activeTriggersPerAtom.Add(atom, state);
				// Don't load changes for a new atom
				loadMode = 2;
			}

			Dictionary<string, List<DAZMorph>> stateMorphBank = null;
			if (state.morphBank == null)
            {
				state.BuildMorphBank(autoLoadTriggerPresets, geometry);
			}
			stateMorphBank = state.morphBank;

			List<AutoLoadTriggerInfo> triggersToUnload = null;
			List<AutoLoadTriggerInfo> triggersToLoad = null;
			HashSet<string> prioritizedMorphTriggers = new HashSet<string>();

			foreach (AutoLoadTriggerInfo loadTrigger in autoLoadTriggerPresets)
			{
				// Find out if this trigger should be currently ACTIVE or NOT.
				bool triggerActive = true;
				// first check personUid.
				triggerActive = (loadTrigger.personUid == "" || loadTrigger.personUid == atom.uid);

				// ANY_MORPH means the trigger is ALWAYS active no matter the person's morphs and can't be overriden by prioritized morphs.

				// check if any prioritized trigger has been found with the same morphName and a personUid already, any triggers after with no personUid but same morph should be ignored.
				if (loadTrigger.personUid == "" && loadTrigger.morphName != ANY_MORPH && enablePrioritizedLoadingTriggers.val && prioritizedMorphTriggers.Contains(loadTrigger.morphName))
				{
					triggerActive = false;
				}

				// continue to check morphs
				if (triggerActive && loadTrigger.morphName != "")
                {
					if (loadTrigger.morphName != ANY_MORPH)
					{
						List<DAZMorph> morphs = null;
						triggerActive = stateMorphBank.TryGetValue(loadTrigger.morphName, out morphs) && morphs.Any((y) => { return y.active; });
						if (!triggerActive)
                        {
							// also check morphBank1.GetMorphByDisplayName in case of morphs in a non-common path (vac/var/scene folder/ etc..)
							DAZMorph morph = geometry.morphBank1.GetMorphByDisplayName(loadTrigger.morphName);
							triggerActive = (morph != null && morph.active);
							if (morphs == null && triggerActive)
                            {
								// need to rebuild the morph bank, look have active morphs not yet in the custom morph bank.
								// This can happen if you load a new look containing morphs inside .var packages for example.
								state.ClearMorphBank();
                            }
						}

						// Old code
						// DAZMorph morph = geometry.morphBank1.GetMorphByDisplayName(morphName);
						// bool morphActive = (morph != null && morph.active);
						// triggerActive = morphActive;
					}
				}

				if (triggerActive && loadTrigger.personUid != "" && loadTrigger.morphName != "")
                {
					prioritizedMorphTriggers.Add(loadTrigger.morphName);
				}

				// Track active triggers for person, any that needs to be loaded/unloaded!?
				if (!triggerActive && state.activeTriggers.Contains(loadTrigger))
				{
					state.activeTriggers.Remove(loadTrigger);
					if (smartRemovalOfPresets.val)
					{
						if (triggersToUnload == null)
							triggersToUnload = new List<AutoLoadTriggerInfo>();
						triggersToUnload.Add(loadTrigger);
					}
				} 
				
				if (triggerActive && !state.activeTriggers.Contains(loadTrigger))
				{
					state.activeTriggers.Add(loadTrigger);
					
					if (triggersToLoad == null)
						triggersToLoad = new List<AutoLoadTriggerInfo>();
					triggersToLoad.Add(loadTrigger);

				}
			}

			// Activate triggers
			if (triggersToLoad != null || triggersToUnload != null)
			{
				//SuperController.LogMessage("PersonUid/Morph change detected!");
				//
				//string debug = "";
				//if (triggersToLoad != null)
				//{
				//	debug += (" Presets to LOAD: " + string.Join(", ", triggersToLoad.Select((x) => { return x.presetName; }).ToArray()));
				//}
				//if (triggersToUnload != null)
				//{
				//	debug += (" Presets to UNLOAD: " + string.Join(", ", triggersToUnload.Select((x) => { return x.presetName; }).ToArray()));
				//}
				//if (debug != "")
				//{
				//	SuperController.LogMessage(debug);
				//}

				core.AutoLoaderTriggered(atom, loadMode,
										 triggersToLoad?.Select((x) => { return x.presetName; }).ToList(), 
										 triggersToUnload?.Select((x) => { return x.presetName; }).ToList());
			}
		}

		protected void SelectedOptionsToggleChanged(bool val)
		{
			SaveOptions();
		}

		public void SaveOptions()
		{ 
			StringBuilder sb = new StringBuilder();
			sb.AppendLine("{");
			sb.AppendLine("  \"id\" : \"Options\",");
			sb.AppendLine("  \"autoSelectPerson\" : \"" + (autoSelectPersonFromScene.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"autoSelectAssetController\" : \"" + (autoSelectAssetController.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"delayPhysicsProperties\" : \"" + (delayPhysicsProperties.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"enableScalingModifier\" : \"" + (enableScalingModifier.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"enableLoadTriggers\" : \"" + (enableLoadTriggers.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"ignoreAllSceneAssets\" : \"" + (ignoreAllSceneAssets.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"ignoreSceneAsssetsContainingScripts\" : \"" + (ignoreSceneAssetsContainingScripts.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"ignoreSceneAssetsMarkedHidden\" : \"" + (ignoreSceneAssetsMarkedHidden.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"smartRemovalOfPresets\" : \"" + (smartRemovalOfPresets.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"enablePersonUidLoadTriggers\" : \"" + (enablePersonUidLoadTriggers.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"updateScriptAtomUids\" : \"" + (updateScriptAtomUids.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"enablePrioritizedLoadingTriggers\" : \"" + (enablePrioritizedLoadingTriggers.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"enablePlayModeUpdates\" : \"" + (enablePlayModeUpdates.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"includedAtomTypes\" : \"" + (string.Join(",", includedAtomTypes.ToArray())) + "\",");
			sb.AppendLine("  \"autoMorphLoaders\" : [");
			AutoLoadTriggerInfo[] loadTriggers = autoLoadTriggerPresets.Where((x) => { return !x.external; }).ToArray();
			for (int i = 0; i < loadTriggers.Length; i++)
			{
				sb.AppendLine("  {");
				sb.AppendLine("  \"id\" : \"" + loadTriggers[i].morphName + "\",");
				sb.AppendLine("  \"personUid\" : \"" + loadTriggers[i].personUid + "\",");
				sb.AppendLine("  \"preset\" : \"" + loadTriggers[i].presetName + "\",");
				sb.AppendLine("  }" + (i < loadTriggers.Length - 1 ? "," : ""));
			}
			sb.AppendLine("  ]");
			sb.AppendLine("}");
			JSONClass mainjc = JSONNode.Parse(sb.ToString()).AsObject;
			try
			{
				SuperController.singleton.SaveJSON(mainjc, core.pluginPath.PluginPresetPath + "/options.json");
			}
			catch (Exception e)
			{
				SuperController.LogError(e.ToString());
			}
		}

		public void LoadOptions()
        {
			// Last read existing plugin options (if any)
			try
			{
				// default:
				includedAtomTypes.Clear();
				autoLoadTriggerPresets.Clear();

				string[] files = SuperController.singleton.GetFilesAtPath(core.pluginPath.PluginPresetPath, "options.json");
				if (files.Length > 0)
				{
					JSONNode jsoptions = SuperController.singleton.LoadJSON(core.pluginPath.PluginPresetPath + "/options.json");

					autoSelectPersonFromScene.valNoCallback = (jsoptions["autoSelectPerson"] == null || jsoptions["autoSelectPerson"].Value == "true");
					autoSelectAssetController.valNoCallback = (jsoptions["autoSelectAssetController"] == null || jsoptions["autoSelectAssetController"].Value == "true");
					delayPhysicsProperties.valNoCallback = (jsoptions["delayPhysicsProperties"] == null || jsoptions["delayPhysicsProperties"].Value == "true");
					enableScalingModifier.valNoCallback = (jsoptions["enableScalingModifier"] == null || jsoptions["enableScalingModifier"].Value == "true");
					enableLoadTriggers.valNoCallback = (jsoptions["enableLoadTriggers"] == null || jsoptions["enableLoadTriggers"].Value == "true");
					ignoreAllSceneAssets.valNoCallback = (jsoptions["ignoreAllSceneAssets"] == null || jsoptions["ignoreAllSceneAssets"].Value == "true");
					ignoreSceneAssetsContainingScripts.valNoCallback = (jsoptions["ignoreSceneAsssetsContainingScripts"] == null || jsoptions["ignoreSceneAsssetsContainingScripts"].Value == "true");
					ignoreSceneAssetsMarkedHidden.valNoCallback = (jsoptions["ignoreSceneAssetsMarkedHidden"] == null || jsoptions["ignoreSceneAssetsMarkedHidden"].Value == "true");
					smartRemovalOfPresets.valNoCallback = (jsoptions["smartRemovalOfPresets"] == null || jsoptions["smartRemovalOfPresets"].Value == "true");
					enablePersonUidLoadTriggers.valNoCallback = (jsoptions["enablePersonUidLoadTriggers"] == null || jsoptions["enablePersonUidLoadTriggers"].Value == "true");
					updateScriptAtomUids.valNoCallback = (jsoptions["updateScriptAtomUids"] == null || jsoptions["updateScriptAtomUids"].Value == "true");
					enablePrioritizedLoadingTriggers.valNoCallback = (jsoptions["enablePrioritizedLoadingTriggers"] != null && jsoptions["enablePrioritizedLoadingTriggers"].Value == "true");
					enablePlayModeUpdates.valNoCallback = (jsoptions["enablePlayModeUpdates"] != null && jsoptions["enablePlayModeUpdates"].Value == "true");

					if (jsoptions["includedAtomTypes"] != null)
					{
						string values = jsoptions["includedAtomTypes"].Value;
						string[] atomTypes = values.Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries);
						includedAtomTypes.AddRange(atomTypes);
					}

					autoLoadTriggerPresets.Clear();
					if (jsoptions["autoMorphLoaders"] != null)
					{
						foreach (JSONNode n in jsoptions["autoMorphLoaders"].AsArray.Childs)
						{
							AutoLoadTriggerInfo loadTrigger = new AutoLoadTriggerInfo(n["personUid"], n["id"], n["preset"], "");
							autoLoadTriggerPresets.Add(loadTrigger);
							// SuperController.LogMessage(loadTrigger.ToString());
						}
					}
				}

				// load external autoMorphLoaders option files
				files = SuperController.singleton.GetFilesAtPath(core.pluginPath.PluginPresetPath, "*.json");
				if (files.Length > 0)
                {
					foreach(string file in files)
                    {
						string fileName = Path.GetFileNameAndExt(file);
						if (!fileName.StartsWith("options_", StringComparison.InvariantCultureIgnoreCase))
                        {
							continue;
                        }

						try
                        {
							JSONNode jsoptions = SuperController.singleton.LoadJSON(core.pluginPath.PluginPresetPath + "/" + fileName);
							if (jsoptions["includedAtomTypes"] != null)
							{
								string values = jsoptions["includedAtomTypes"].Value;
								string[] atomTypes = values.Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries);
								includedAtomTypes.AddRange(atomTypes);
							}

							if (jsoptions["autoMorphLoaders"] != null)
							{
								foreach (JSONNode n in jsoptions["autoMorphLoaders"].AsArray.Childs)
								{
									AutoLoadTriggerInfo loadTrigger = new AutoLoadTriggerInfo(n["personUid"], n["id"], n["preset"], fileName);
									autoLoadTriggerPresets.Add(loadTrigger);
									// SuperController.LogMessage(loadTrigger.ToString());
								}
							}
						}
						catch (Exception e)
						{
							// Ignore, file might not exist or loading values failed...
							SuperController.LogError("");
							SuperController.LogError("Failed to load option file " + file);
							SuperController.LogError(e.ToString());
						}
					}
				}

			}
			catch (Exception e)
			{
				// Ignore, file might not exist or loading values failed...
				SuperController.LogError(e.ToString());
			}

			// Remove duplicates from includedAtomTypes
			includedAtomTypes = includedAtomTypes.Distinct().ToList(); // easy way of doing it, will work!
			// Force "CustomUnityAsset" to be included, plugin designed for that
			if (!includedAtomTypes.Contains("CustomUnityAsset"))
            {
				includedAtomTypes.Insert(0, "CustomUnityAsset");
			}

			// Remove duplicates from autoLoaderMorphPresets
			HashSet<string> keys = new HashSet<string>();
			for (int i = 0; i < autoLoadTriggerPresets.Count; i++)
			{
				if (!keys.Add(autoLoadTriggerPresets[i].presetName + ":" + autoLoadTriggerPresets[i].morphName + ":" + autoLoadTriggerPresets[i].presetName))
                {
					// duplicate
					// SuperController.LogMessage("Removing duplicate: " + autoLoaderMorphPresets[i].ToString());
					autoLoadTriggerPresets.RemoveAt(i);
					i--;
				}
			}
			SortAutoLoadTriggerPresets();

			UpdateAutoLoaderPresetsList();
			UpdateIncludedAtomTypesList();
		}

		private void SortAutoLoadTriggerPresets()
        {
			// let personUid triggers come before others, this will be important to make certain assumtions in the auto-loader routine check optimized
			autoLoadTriggerPresets.Sort((a, b) => {
				int c = string.Compare(b.personUid, a.personUid);
				if (c == 0)
                {
					return string.Compare(a.morphName, b.morphName);
                }
				return c;
			});
			// SuperController.LogMessage("autoLoadTriggerPresets: " + string.Join(", ", autoLoadTriggerPresets.Select((x) => { return x.personUid + ":" + x.morphName; }).ToArray()));
		}

		public void UpdateAutoLoaderPresetsList()
		{
			autoLoaderPresetsList.val = "Auto load preset triggers:\n\n" + string.Join("\n\n", autoLoadTriggerPresets.Select((x, i) => {
				return "Trigger " + i.ToString() + ":\n" + (x.personUid != "" ? "personUid: " + x.personUid+ "\n" : "") + (x.morphName != "" ? "morph: " + x.morphName + "\n" : "") + "preset: " + x.presetName + (!x.external ? "" : "\nfile: " + x.externalFile); 
			}).ToArray());
		}
		public void UpdateIncludedAtomTypesList()
        {
			includedAtomTypesList.val = "Atom types to handle (assets):\n\n" + string.Join("\n", includedAtomTypes.ToArray());
        }

		public void AddLoadTrigger(string personUid, string morphDisplayName, string presetName)
		{
			if (personUid == null)
				personUid = "";
			if (morphDisplayName == null)
				morphDisplayName = "";
			
			if (presetName == null || presetName == null || personUid == "" && morphDisplayName == "")
            {
				// nothing to add
				return;
            }
			
			// Don't add duplicates.
			if (autoLoadTriggerPresets.FirstOrDefault((x) => { return x.personUid == personUid && x.morphName == morphDisplayName && x.presetName == presetName; }) != null)
			{
				SuperController.LogMessage("Trigger already added. Skipping duplicate.");
			}
			else
			{
				// Log("Trigger added.");
				autoLoadTriggerPresets.Add(new AutoLoadTriggerInfo(personUid, morphDisplayName, presetName, ""));
				SortAutoLoadTriggerPresets();
				SelectedOptionsToggleChanged(true); // trigger save options

				// Update any active triggers
				foreach (KeyValuePair<Atom, AtomTriggerStates> pair in activeTriggersPerAtom)
				{
					pair.Value.ClearMorphBank();
				}

				// Force trigger unload on the newly saved trigger, as it soon will be triggered anyway!
				if (enableLoadTriggers.val && core.rootAtom != null && !core.rootAtom.destroyed)
				{
					core.AutoLoaderTriggered(core.rootAtom, 1, null, new List<string>(new string[] { presetName }));
				}

			}
			UpdateAutoLoaderPresetsList();
		}

		public void RemoveLoadTriggerName(string name)
		{
			int index = GetLoadTriggerNames().IndexOf(name);
			if (index >= 0)
			{
				AutoLoadTriggerInfo loadTrigger = autoLoadTriggerPresets[index];
				if (loadTrigger.external)
				{
					SuperController.LogError("This trigger is loaded from an external options file (" + loadTrigger.externalFile + "). Triggers loaded from external option files must be removed manually from the plugin preset folder and plugin reloaded.");
				}
				else
				{
					autoLoadTriggerPresets.RemoveAt(index);
					SortAutoLoadTriggerPresets();
					SelectedOptionsToggleChanged(true); // trigger save options

					// Update any active triggers
					foreach (KeyValuePair<Atom, AtomTriggerStates> pair in activeTriggersPerAtom)
					{
						pair.Value.activeTriggers.Remove(loadTrigger);
						pair.Value.ClearMorphBank();
					}
					
					// Force trigger unload on the newly saved trigger, as it soon will be triggered anyway!
					if (enableLoadTriggers.val && core.rootAtom != null && !core.rootAtom.destroyed)
					{
						core.AutoLoaderTriggered(core.rootAtom, 2, null, new List<string>(new string[] { loadTrigger.presetName }));
					}
				}
			}
			UpdateAutoLoaderPresetsList();
		}

		public List<string> GetPersonUidTriggerNames()
        {
			List<string> result = new List<string>();
			result.Add(""); // empty selection first!
			if (core.rootAtom != null && !core.rootAtom.destroyed)
            {
				result.Add(core.rootAtom.uid);
            }
			return result;
		}

		public List<string> GetMorphTriggerNames()
		{
			List<string> result = new List<string>();
			if (core.rootAtom != null && !core.rootAtom.destroyed)
			{
				DAZCharacterSelector geometry = core.rootAtom.GetStorableByID("geometry") as DAZCharacterSelector;
				if (geometry != null)
				{
					foreach (DAZMorph morph in geometry.morphBank1.morphs)
					{
						if (morph != null && morph.visible && morph.active && (morph.startValue != morph.appliedValue || morph.startValue != morph.morphValue))
						{
							result.Add(morph.displayName);
						}
					}
				}
			}
			result.Sort();
			result.Insert(0, ""); // empty selection first!
			result.Add(ANY_MORPH); // wildcard selection last (this will trigger on any person!)
			return result;
		}

		public List<string> GetLoadTriggerNames()
		{
			List<string> triggerNames = new List<string>();
			for (int i = 0; i < autoLoadTriggerPresets.Count; i++)
			{
				// build a short text that fits in the popup
				triggerNames.Add(string.Format("Trigger {0}", i.ToString()));
			}
			return triggerNames;
		}

		public bool CheckIncludeAtomType(string atomType)
        {
			// Checks if the atom should be included to be handled by this script.
			if (includedAtomTypes.Contains(atomType))
            {
				return true;
            }
			return false;
        }

		public List<string> GetIncludedAtomTypes()
        {
			return new List<string>(includedAtomTypes);
        }

		public List<string> GetPossibleLinkTypes()
        {
			return new List<string>(new string[]{
				FreeControllerV3.PositionState.On.ToString(),
				FreeControllerV3.PositionState.Comply.ToString(),
				FreeControllerV3.PositionState.Off.ToString(),
				FreeControllerV3.PositionState.ParentLink.ToString(),
				FreeControllerV3.PositionState.PhysicsLink.ToString(),
				FreeControllerV3.PositionState.Hold.ToString(),
				FreeControllerV3.PositionState.Lock.ToString()
			});
		}

		public string GetAtomLinkType(Atom apAtom)
        {
			if (apAtom.mainController.currentPositionState.ToString() == apAtom.mainController.currentRotationState.ToString())
            {
				return apAtom.mainController.currentPositionState.ToString();
			}
			// A mixed combination
			return apAtom.mainController.currentPositionState.ToString() + " / " + apAtom.mainController.currentRotationState.ToString();
		}

		public FreeControllerV3.PositionState GetAtomLinkTypePosition(string linkType, FreeControllerV3.PositionState defaultValue)
        {
			switch(linkType)
            {
				case "On": return FreeControllerV3.PositionState.On;
				case "Comply": return FreeControllerV3.PositionState.Comply;
				case "Off": return FreeControllerV3.PositionState.Off;
				case "ParentLink": return FreeControllerV3.PositionState.ParentLink;
				case "PhysicsLink": return FreeControllerV3.PositionState.PhysicsLink;
				case "Hold": return FreeControllerV3.PositionState.Hold;
				case "Lock": return FreeControllerV3.PositionState.Lock;
			}
			return defaultValue;
		}

		public FreeControllerV3.RotationState GetAtomLinkTypeRotation(string linkType, FreeControllerV3.RotationState defaultValue)
		{
			switch (linkType)
			{
				case "On": return FreeControllerV3.RotationState.On;
				case "Comply": return FreeControllerV3.RotationState.Comply;
				case "Off": return FreeControllerV3.RotationState.Off;
				case "ParentLink": return FreeControllerV3.RotationState.ParentLink;
				case "PhysicsLink": return FreeControllerV3.RotationState.PhysicsLink;
				case "Hold": return FreeControllerV3.RotationState.Hold;
				case "Lock": return FreeControllerV3.RotationState.Lock;
			}
			return defaultValue;
		}

		public bool TrackLinkInWorldSpace(FreeControllerV3.PositionState state)
        {
			return false; // (state != FreeControllerV3.PositionState.ParentLink && state != FreeControllerV3.PositionState.PhysicsLink);
		}

		public bool TrackLinkInWorldSpace(FreeControllerV3.RotationState state)
		{
			return false; // (state != FreeControllerV3.RotationState.ParentLink && state != FreeControllerV3.RotationState.PhysicsLink);
		}

		public bool LiveUpdateLink(FreeControllerV3.PositionState state)
		{
			return (state != FreeControllerV3.PositionState.ParentLink && state != FreeControllerV3.PositionState.PhysicsLink);
		}

		public bool LiveUpdateLink(FreeControllerV3.RotationState state)
		{
			return (state != FreeControllerV3.RotationState.ParentLink && state != FreeControllerV3.RotationState.PhysicsLink);
		}


	}

	public class AtomTriggerStates
    {
		public List<AutoLoadTriggerInfo> activeTriggers;
		/// <summary>
        /// Morph bank containing morph displayName and all morphs using that displayName. This plugin as a whole operates on morph displayNames...
        /// </summary>
		public Dictionary<string, List<DAZMorph>> morphBank;
		public AtomTriggerStates()
        {
			activeTriggers = new List<AutoLoadTriggerInfo>();
			ClearMorphBank();
		}

		public void ClearMorphBank()
        {
			morphBank = null;
        }

		/// <summary>
        /// Builds up a custom morphBank over morphs bound to load triggers to keep the morphBank small and to be able to have multiple morphs with the same displayName.
        /// </summary>
        /// <param name="autoLoadTriggerPresets"></param>
        /// <param name="geometry"></param>
		public void BuildMorphBank(List<AutoLoadTriggerInfo> autoLoadTriggerPresets, DAZCharacterSelector geometry)
        {
			// SuperController.LogMessage("Building morph bank");
			// rebuild a private morphbank over all morph that's bound to a trigger in this plugin to keep data lookup small.
			morphBank = new Dictionary<string, List<DAZMorph>>();
			List<DAZMorph> morphs = null;
			foreach (AutoLoadTriggerInfo loadTrigger in autoLoadTriggerPresets)
			{
				if (loadTrigger.morphName != Options.ANY_MORPH)
				{
					foreach (DAZMorph morph in geometry.morphBank1.morphs.Where((x) => { return x.displayName == loadTrigger.morphName; }))
					{
						if (!morphBank.TryGetValue(morph.displayName, out morphs))
						{
							morphs = new List<DAZMorph>();
							morphBank.Add(morph.displayName, morphs);
						}
						morphs.Add(morph);
					}
				}
			}
		}
	}

	public class AutoLoadTriggerInfo
    {
		public string personUid;
		public string morphName;
		public string presetName;
		public bool external;
		public string externalFile;

		public AutoLoadTriggerInfo(string uid, string morph, string preset, string extFile)
        {
			personUid = (uid == null ? "" : uid);
			morphName = (morph == null ? "" : morph);
			presetName = (preset == null ? "" : preset);
			externalFile = (extFile == null ? "" : extFile);
			external = (externalFile != "");
		}

        public override string ToString()
        {
			return string.Format("{0}:{1}:{2}:{3}:{4}", personUid, morphName, presetName, external.ToString(), externalFile);
        }

    }

}